/************************************************************************
 * libidmef: A library to create IDMEF messages in XML format.
 * Author: Joe McAlerney, Silicon Defense, (joey@SiliconDefense.com)
 *
 * Copyright (c) 2000,2001 by Silicon Defense (http://www.silicondefense.com/)
 * 
 * This library is released under the GNU GPL and BSD software licenses.
 * You may choose to use one or the other, BUT NOT BOTH.  The GNU GPL
 * license is located in the file named COPYING.  The BSD license is located
 * in the file named COPYING.BSD.  Please contact us if there are any
 * questions.
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "time.h"

#include "idmefconfig.h"
#include "idmefxml.h"


/*******************************************
 * Global variable definitions
 *******************************************/

xmlDocPtr doc;
xmlDtdPtr dtd;
xmlValidCtxtPtr vctxt;
char *g_dtd_path;
xmlOutputBufferPtr g_output;


/*************************************************************************
 * With the exception of newSimpleElement, each of these "new" functions 
 * accepts an xmlNodePtr, and a variable list of parameters, which should be
 * xmlNodePtrs as well.  The parameters must be passed in the order specified
 * by the DTD.  The functions will check to see if all required parameters 
 * were passed.  If not, it will return NULL.  Possible parameters are
 * displayed above the functions.
 *
 * Notes: At least ONE parameter is required in each function.  This parameter
 *        is labled 'first'. 
 *
 *        You must pass NULL in as the last argument. This is to let the 
 *        function know when to quit reading in arguments.  Failure to do this
 *        will result in "undefined behavior".
 *************************************************************************/
 
/*********************************************************************
 *********************************************************************
 *****								 *****
 *****			The Message Classes			 *****
 *****								 *****
 *********************************************************************
 *********************************************************************/

/*==================================================================
 * Name: newIDMEF_Message
 *
 * Declaration: xmlNodePtr newIDMEF_Message(xmlNodePtr, ...);
 *
 * Possible parameters:
 *
 *   ATTRIBUTES
 *   xmlNodePtr version: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr Alert    : zero or more
 *   xmlNodePtr Heartbeat: zero or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the root element
 *
 * Example:
 *	    if(newIDMEF_Message(theAlert, NULL))
 *              printCurrentMessage(stdout);
 *===================================================================*/

xmlNodePtr newIDMEF_Message(xmlNodePtr first, ...)
{
  xmlNodePtr p, theRoot;
  va_list ap;
  xmlChar *tmp_content = NULL;


  if(doc == NULL)        /* have we called initialize successfully yet? */
  {
      fprintf(stderr,"idmefxml.c: newIDMEF_Message(): you must call createCurrentDoc first\n");
      return NULL;
  }

  theRoot = xmlDocGetRootElement(doc);
  
  /* Set the IDMEF Version attribute */
  xmlSetProp(theRoot, "version", IDMEF_MESSAGE_VERSION);

  if(first == NULL) /* do we have at least one parameter? */
    return theRoot;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);
       
     switch(*(p->name))
     {
          case 'v':
          if(strcmp(p->name, "version") == 0)
	  {   
 	     xmlSetProp(theRoot, "version", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"IDMEF-Message");
          break;
        case 'A':
          if(strcmp(p->name, "Alert") == 0)
  	    xmlAddChild(theRoot, p);
          else
            badNode(p,"IDMEF-Message");
          break;
        case 'H':
          if(strcmp(p->name, "Heartbeat") == 0)
 	     xmlAddChild(theRoot, p);
          else
            badNode(p,"IDMEF-Message");
          break;
        default:
          badNode(p,"IDMEF-Message");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(theRoot);
}

/*==================================================================
 * Name: newAlert
 *
 * Declaration: xmlNodePtr newAlert(xmlNodePtr, ...);
 *
 * Possible parameters:
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident  : optional
 +
 *   SUB-ELEMENTS
 *   xmlNodePtr Analyzer        : exactly one
 *   xmlNodePtr CreateTime      : exactly one
 *   xmlNodePtr DetectTime      : zero or one
 *   xmlNodePtr AnalyzerTime    : zero or one
 *   xmlNodePtr Source          : zero or more
 *   xmlNodePtr Target          : zero or more
 *   xmlNodePtr Classification  : one or more
 +   xmlNodePtr Assessment      : zero or one
 *   xmlNodePtr ToolAlert       : zero or one
 *   xmlNodePtr OverflowAlert   : zero or one
 *   xmlNodePtr CorrelationAlert: zero or one
 *   xmlNodePtr AdditionalData  : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Alert node.
 *
 * Example:
 *	    xmlNodePtr theAlert = newAlert(myidentd, myCreateTime,
 *                                         myAnalyzer,myClassification,
 *                                         mySource,myTarget, NULL);
 *
 *===================================================================*/

xmlNodePtr newAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Alert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAlert(): xmlNewNode(NULL,\"Alert\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;
       
     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i':
	  if(strcmp(p->name, "ident") == 0)
	  {
	      xmlSetProp(cur, "ident", tmp_content);
	      xmlFreeNode(p);
	   }         
	   else
	      badNode(p,"Alert");
           break;
        case 'T':
          switch(*(p->name + 1))
	  {
   	     case 'a':
               if(strcmp(p->name, "Target") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'o':
               if(strcmp(p->name, "ToolAlert") == 0)
	       {
   	         /* check to see if an alert subclass already exists */
		 if(hasElement(cur,"CorrelationAlert") ||
                    hasElement(cur,"OverflowAlert") ||
                    hasElement(cur,"ToolAlert"))
		 {
		   fprintf(stderr,"idmefxml.c: DTD Violation - Parent Alert already contains a CorrelationAlert, OverflowAlert, or ToolAlert\n");
                   xmlFree(p);
                 }
                 else
  	          xmlAddChild(cur, p);
               }
               else
                  badNode(p,"Alert");
               break;
  	     default:
               badNode(p,"Alert");
	  }
          break;
        case 'A':
          switch(*(p->name + 1))
	  {
	     case 'n':   /* Analyzer or AnalyzerTime */
               if((strcmp(p->name, "Analyzer") == 0) ||
		  (strcmp(p->name, "AnalyzerTime") == 0)) 
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'd':
               if(strcmp(p->name, "AdditionalData") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 's':
               if(strcmp(p->name, "Assessment") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
    	     default:
               badNode(p,"Alert");   
          }
          break;
        case 'C':
          switch(*(p->name + 1))
	  {
     	     case 'l':
               if(strcmp(p->name, "Classification") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'o':
               if(strcmp(p->name, "CorrelationAlert") == 0)
	       {
   	         /* check to see if an alert subclass already exists */
		 if(hasElement(cur,"CorrelationAlert") ||
                    hasElement(cur,"OverflowAlert") ||
                    hasElement(cur,"ToolAlert"))
		 {
		   fprintf(stderr,"idmefxml.c: DTD Violation - Parent Alert already contains a CorrelationAlert, OverflowAlert, or ToolAlert\n");
                   xmlFree(p);
                 }
                 else
  	          xmlAddChild(cur, p);
               }
               else
                  badNode(p,"Alert");
               break;
     	     case 'r':
               if(strcmp(p->name, "CreateTime") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
    	     default:
               badNode(p,"Alert");   
          }
          break;
        case 'D':
          if(strcmp(p->name, "DetectTime") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Alert");
          break;
        case 'S':
          if(strcmp(p->name, "Source") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Alert");
          break;
        case 'O':
          if(strcmp(p->name, "OverflowAlert") == 0)
	  {
 	     /* check to see if an alert subclass already exists */
	     if(hasElement(cur,"CorrelationAlert") ||
                hasElement(cur,"OverflowAlert") ||
                hasElement(cur,"ToolAlert"))
	     {
	       fprintf(stderr,"idmefxml.c: DTD Violation - Parent Alert already contains a CorrelationAlert, OverflowAlert, or ToolAlert\n");
               xmlFree(p);
             }
             else
               xmlAddChild(cur, p);
          }
          else
            badNode(p,"Alert");
          break;
        default:
          badNode(p,"Alert");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newToolAlert
 *
 * Declaration: xmlNodePtr newToolAlert(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr name      : exactly one
 *   xmlNodePtr command   : zero or more
 *   xmlNodePtr alertident: one or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newToolAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;

  cur = xmlNewNode(NULL,"ToolAlert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newToolAlert: xmlNewNode(NULL,\"ToolAlert\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     switch(*(p->name))
     {
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"ToolAlert");
          break;
        case 'c': /* command */
          if(strcmp(p->name, "command") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"ToolAlert");
          break;
        case 'a': /* alertident */
          if(strcmp(p->name, "alertident") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"ToolAlert");
          break;
        default:
          badNode(p,"ToolAlert");
     }
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newCorrelationAlert
 *
 * Declaration: xmlNodePtr newCorrelationAlert(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr name      : exactly one
 *   xmlNodePtr alertident: one or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created CorrelationAlert node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newCorrelationAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;

  cur = xmlNewNode(NULL,"CorrelationAlert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newCorrelationAlert: xmlNewNode(NULL,\"CorrelationAlert\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     switch(*(p->name))
     {
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"CorrelationAlert");
          break;
        case 'a': /* alertident */
          if(strcmp(p->name, "alertident") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"CorrelationAlert");
          break;
        default:
          badNode(p,"CorrelationAlert");
     }
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newOverflowAlert
 *
 * Declaration: xmlNodePtr newOverflowAlert(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr program: exactly one
 *   xmlNodePtr size   : zero or one
 *   xmlNodePtr buffer : zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created OverflowAlert node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newOverflowAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;

  cur = xmlNewNode(NULL,"OverflowAlert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newOverflowAlert: xmlNewNode(NULL,\"OverflowAlert\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     switch(*(p->name))
     {
        case 'p': /* program */
          if(strcmp(p->name, "program") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"OverflowAlert");
          break;
        case 's': /* size */
          if(strcmp(p->name, "size") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"OverflowAlert");
          break;
        case 'b': /* buffer */
          if(strcmp(p->name, "buffer") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"OverflowAlert");
          break;
        default:
          badNode(p,"OverflowAlert");
     }
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newHeartbeat
 *
 * Declaration: xmlNodePtr newHeartbeat(xmlNodePtr, ...);
 *
 * Possible parameters:
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr Analyzer        : exactly one
 *   xmlNodePtr CreateTime      : exactly one
 *   xmlNodePtr AnalyzerTime    : zero or one
 *   xmlNodePtr AdditionalData  : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Heartbeat node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newHeartbeat(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Heartbeat");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newHeartbeat(): xmlNewNode(NULL,\"Heartbeat\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Heartbeat");
          break;
        case 'C': /* CreateTime */
          if(strcmp(p->name, "CreateTime") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Heartbeat");
          break;
        case 'A':
          switch(*(p->name + 1))
	  {
	     case 'n':   /* Analyzer or AnalyzerTime */
               if((strcmp(p->name, "Analyzer") == 0) ||
		  (strcmp(p->name, "AnalyzerTime") == 0)) 
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'd':
               if(strcmp(p->name, "AdditionalData") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
    	     default:
               badNode(p,"Alert");   
          }
          break;
        default:
          badNode(p,"Heartbeat");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*********************************************************************
 *********************************************************************
 *****								 *****
 *****			The Core Classes			 *****
 *****								 *****
 *********************************************************************
 *********************************************************************/

/*==================================================================
 * Name: newAnalyzer
 *
 * Declaration: xmlNodePtr newAnalyzer(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr analyzerid  : optional (but see spec)
 *   xmlNodePtr manufacturer: optional
 *   xmlNodePtr model       : optional
 *   xmlNodePtr version     : optional
 *   xmlNodePtr class       : optional
 +   xmlNodePtr ostype      : optional
 +   xmlNodePtr osversion   : optional
 +
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr Node   : zero or one
 *   xmlNodePtr Process: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Analyzer node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newAnalyzer(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Analyzer");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAnalyzer: xmlNewNode(NULL,\"Analyzer\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'a': /* analyzerid */
          if(strcmp(p->name, "analyzerid") == 0)
	  {
 	     xmlSetProp(cur, "analyzerid", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Analyzer");
          break;
        case 'c': /* class */
          if(strcmp(p->name, "class") == 0)
	  {
 	     xmlSetProp(cur, "class", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Analyzer");
          break;
        case 'm':
          switch(*(p->name + 1))
	  {
 	     case 'a':   /* manufacturer */
               if((strcmp(p->name, "manufacturer") == 0)) 
    	          xmlSetProp(cur, "manufacturer", tmp_content);
               else
                  badNode(p,"Analyzer");
               break;
	     case 'o':   /* model */
               if(strcmp(p->name, "model") == 0)
    	          xmlSetProp(cur, "model", tmp_content);
               else
                  badNode(p,"Analyzer");
               break;
    	     default:
               badNode(p,"Analyzer");   
          }
          break;
        case 'o':
          switch(*(p->name + 2))
	  {
 	     case 't':   /* ostype */
               if((strcmp(p->name, "ostype") == 0)) 
    	          xmlSetProp(cur, "ostype", tmp_content);
               else
                  badNode(p,"Analyzer");
               break;
	     case 'v':   /* osversion */
               if(strcmp(p->name, "osversion") == 0)
    	          xmlSetProp(cur, "osversion", tmp_content);
               else
                  badNode(p,"Analyzer");
               break;
    	     default:
               badNode(p,"Analyzer");   
          }
          break;
        case 'v': /* version */
          if(strcmp(p->name, "version") == 0)
	  {
 	     xmlSetProp(cur, "version", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Analyzer");
          break;
        case 'N': /* Node */
	  if(strcmp(p->name, "Node") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Analyzer");
          break;
        case 'P': /* Process */
          if(strcmp(p->name, "Process") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Analyzer");
          break;
        default:
          badNode(p,"Analyzer");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newClassification
 *
 * Declaration: xmlNodePtr newClassification(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr origin: required
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr name: exactly one
 *   xmlNodePtr url : exactly one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Classification node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newClassification(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Classification");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newClassification: xmlNewNode(NULL,\"Classification\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'o': /* origin */
          if(strcmp(p->name, "origin") == 0)
	  {
 	     xmlSetProp(cur, "origin", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Classification");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Classification");
          break;
        case 'u': /* url */
          if(strcmp(p->name, "url") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Classification");
          break;
        default:
          badNode(p,"Classification");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newSource
 *
 * Declaration: xmlNodePtr newSource(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident  : optional
 *   xmlNodePtr spoofed: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr Node   : zero or one
 *   xmlNodePtr User   : zero or one
 *   xmlNodePtr Process: zero or one
 *   xmlNodePtr Service: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Source node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newSource(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Source");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newSource: xmlNewNode(NULL,\"Source\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
	  if(strcmp(p->name, "ident") == 0)
	  {
	    xmlSetProp(cur, "ident", tmp_content);
	    xmlFreeNode(p);
	  }
	  else
	    badNode(p,"Source");
	  break;
        case 's': /* spoofed */
	  if(strcmp(p->name, "spoofed") == 0)
	  {
	    xmlSetProp(cur, "spoofed", tmp_content);
	    xmlFreeNode(p);
	  }
	  else
	    badNode(p,"Source");
	  break;
        case 'N': /* Node */
	  if(strcmp(p->name, "Node") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Source");
          break;
        case 'U': /* User */
	  if(strcmp(p->name, "User") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Source");
          break;
        case 'P': /* Process */
          if(strcmp(p->name, "Process") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Source");
          break;
        case 'S': /* Service */
          if(strcmp(p->name, "Service") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Source");
          break;
        default:
          badNode(p,"Source");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newTarget
 *
 * Declaration: xmlNodePtr newTarget(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident: optional
 *   xmlNodePtr decoy: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr Node    : zero or one
 *   xmlNodePtr User    : zero or one
 *   xmlNodePtr Process : zero or one
 *   xmlNodePtr Service : zero or one
 *   xmlNodePtr FileList: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Target node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newTarget(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Target");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newTarget: xmlNewNode(NULL,\"Target\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Target");
          break;
	case 'd': /* decoy */
	  if(strcmp(p->name, "decoy") == 0)
	  {
	     xmlSetProp(cur, "decoy", tmp_content);
             xmlFreeNode(p);
          }
	  else
	    badNode(p,"Target");
	  break;
        case 'N': /* Node */
	  if(strcmp(p->name, "Node") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Target");
          break;
        case 'U': /* User */
	  if(strcmp(p->name, "User") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Target");
          break;
        case 'P': /* Process */
          if(strcmp(p->name, "Process") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Target");
          break;
        case 'S': /* Service */
          if(strcmp(p->name, "Service") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Target");
          break;
        case 'F': /* FileList */
          if(strcmp(p->name, "FileList") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Target");
          break;
        default:
          badNode(p,"Target");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAssessment
 *
 * Declaration: xmlNodePtr newAssessment(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr Impact    : zero or one
 *   xmlNodePtr Action    : zero or more
 *   xmlNodePtr Confidence: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Assessment node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newAssessment(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Assessment");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAssessment: xmlNewNode(NULL,\"Assessment\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'I': /* Impact */
	  if(strcmp(p->name, "Impact") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Assessment");
          break;
        case 'A': /* Action */
	  if(strcmp(p->name, "Action") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Assessment");
          break;
        case 'C': /* Confidence */
          if(strcmp(p->name, "Confidence") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Assessment");
          break;
        default:
          badNode(p,"Assessment");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAdditionalData
 *
 * Declaration: xmlNodePtr newAdditionalData(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr type   : required
 *   xmlNodePtr meaning: optional
 *
 *   VALUE
 *   xmlNodePtr value: exactly one (the additional data)
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created AdditionalData node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newAdditionalData(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"AdditionalData");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAdditionalData: xmlNewNode(NULL,\"AdditionalData\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'm': /* meaning */
          if(strcmp(p->name, "meaning") == 0)
	  {
 	     xmlSetProp(cur, "meaning", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AdditionalData");
          break;
        case 't': /* type */
          if(strcmp(p->name, "type") == 0)
	  {
 	     xmlSetProp(cur, "type", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AdditionalData");
          break;
        case 'v': /* value */
          if(strcmp(p->name, "value") == 0)
	  {
  	     xmlNodeSetContent(cur, tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AdditionalData");
          break;
        default:
          badNode(p,"AdditionalData");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);

}

/*********************************************************************
 *********************************************************************
 *****								 *****
 *****			The Time Classes			 *****
 *****								 *****
 *********************************************************************
 *********************************************************************/

/*==================================================================
 * Name: newCreateTime
 *
 * Declaration: xmlNodePtr newCreateTime(struct timeval *);
 *
 * parameter 
 * 
 *   struct timeval * : If provided, the timeval will be used to create the
 *                      ntpstamp and DATETIME.  NULL must be passed in if
 *                      the newCreateTime function is to create the ntpstamp
 *                      and DATETIME itself.
 *
 * Returns: an xmlNodePtr to the newly created CreateTime node.
 *
 * Example: xmlNodePtr createTime =  newCreateTime(NULL);
 *          xmlNodePtr createTime2 = newCreateTime(&mytimeval); // use provided
 *                                                             
 *===================================================================*/

xmlNodePtr newCreateTime(struct timeval *tv_arg)
{
  xmlNodePtr cur;
  struct timeval currtv;
  xmlChar *ntpstamp, *datetime;

  cur = xmlNewNode(NULL,"CreateTime");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newCreateTime: xmlNewNode(NULL,\"CreateTime\") failed\n");
    return NULL;
  }

  /* Create the ntpstamp and datetime strings.  If a timeval argument
     was provided, use it.  Otherwise, create a new one. */

  if(tv_arg != NULL)            /* timeval parameter provided */
    currtv = *tv_arg;
  else                          /* timeval was not provided, get current time */
    gettimeofday(&currtv,0);

  /* grab the timestamp and datetime */
 
  ntpstamp = timevalToNtpTimestamp(&currtv);
  datetime = timevalToDatetime(&currtv);

  /* Set the ntpstamp attribute, and DATETIME value */

  xmlSetProp(cur, "ntpstamp", ntpstamp);
  xmlFree(ntpstamp);

  xmlNodeSetContent(cur, datetime);
  xmlFree(datetime);
  return(cur);
}

/*==================================================================
 * Name: newAnalyzerTime
 *
 * Declaration: xmlNodePtr newAnalyzerTime(struct timeval *);
 *
 * parameter 
 * 
 *   struct timeval * : If provided, the timeval will be used to create the
 *                      ntpstamp and DATETIME.  NULL must be passed in if
 *                      the newAnalyzerTime function is to create the ntpstamp
 *                      and DATETIME itself.
 *
 * Returns: an xmlNodePtr to the newly created AnayzerTime node.
 *
 * Example: xmlNodePtr analyzerTime =  newAnalyzerTime(NULL);
 *          xmlNodePtr analyzerTime2 = newAnalyzerTime(&mytimeval); 
 *                                                             
 *===================================================================*/

xmlNodePtr newAnalyzerTime(struct timeval *tv_arg)
{
  xmlNodePtr cur;
  struct timeval currtv;
  xmlChar *ntpstamp, *datetime;

  cur = xmlNewNode(NULL,"AnalyzerTime");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAnalyzerTime: xmlNewNode(NULL,\"AnalyzerTime\") failed\n");
    return NULL;
  }

  /* Create the ntpstamp and datetime strings.  If a timeval argument
     was provided, use it.  Otherwise, create a new one. */

  if(tv_arg != NULL)            /* timeval parameter provided */
    currtv = *tv_arg;
  else                          /* timeval was not provided, get current time */
    gettimeofday(&currtv,0);

  ntpstamp = timevalToNtpTimestamp(&currtv);
  datetime = timevalToDatetime(&currtv);

  /* Set the ntpstamp attribute, and DATETIME value */

  xmlSetProp(cur, "ntpstamp", ntpstamp);
  xmlFree(ntpstamp);

  xmlNodeSetContent(cur, datetime);
  xmlFree(datetime);

  return(cur);
}

/*==================================================================
 * Name: newDetectTime
 *
 * Declaration: xmlNodePtr newDetectTime(struct timeval *);
 *
 * parameter 
 * 
 *   struct timeval * : If provided, the timeval will be used to create the
 *                      ntpstamp and DATETIME.  NULL must be passed in if
 *                      the newDetectTime function is to create the ntpstamp
 *                      and DATETIME itself.
 *
 * Returns: an xmlNodePtr to the newly created Time node.
 *
 * Example: xmlNodePtr detectTime =  newDetectTime(NULL);
 *          xmlNodePtr detectTime2 = newDetectTime(&mytimeval); // use provided
 *                                                             
 *===================================================================*/

xmlNodePtr newDetectTime(struct timeval *tv_arg)
{
  xmlNodePtr cur;
  struct timeval currtv;
  xmlChar *ntpstamp, *datetime;

  cur = xmlNewNode(NULL,"DetectTime");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newDetectTime: xmlNewNode(NULL,\"DetectTime\") failed\n");
    return NULL;
  }

  /* Create the ntpstamp and datetime strings.  If a timeval argument
     was provided, use it.  Otherwise, create a new one. */

  if(tv_arg != NULL)            /* timeval parameter provided */
    currtv = *tv_arg;
  else                          /* timeval was not provided, get current time */
    gettimeofday(&currtv,0);

  ntpstamp = timevalToNtpTimestamp(&currtv);
  datetime = timevalToDatetime(&currtv);

  /* Set the ntpstamp attribute, and DATETIME value */

  xmlSetProp(cur, "ntpstamp", ntpstamp);
  xmlFree(ntpstamp);

  xmlNodeSetContent(cur, datetime);
  xmlFree(datetime);

  return(cur);
}

/*********************************************************************
 *********************************************************************
 *****								 *****
 *****			The Support Classes			 *****
 *****								 *****
 *********************************************************************
 *********************************************************************/

/*==================================================================
 * Name: newNode
 *
 * Declaration: xmlNodePtr newNode(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident   : optional
 *   xmlNodePtr category: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr location: zero or one
 *   xmlNodePtr name    : zero or one
 *   xmlNodePtr Address : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Node node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newNode(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Node");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newNode: xmlNewNode(NULL,\"Node\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Node");
          break;
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Node");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Node");
          break;
        case 'l': /* location */
          if(strcmp(p->name, "location") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Node");
          break;
        case 'A': /* Address */
          if(strcmp(p->name, "Address") == 0)
            xmlAddChild(cur,p);
          else
            badNode(p,"Node");
          break;
        default:
          badNode(p,"Node");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAddress
 *
 * Declaration: xmlNodePtr newAddress(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident    : optional
 *   xmlNodePtr category : optional
 *   xmlNodePtr vlan-name: optional
 *   xmlNodePtr vlan-num : optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr address: exactly one
 *   xmlNodePtr netmask: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Address node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newAddress(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Address");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAddress: xmlNewNode(NULL,\"Address\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Address");
          break;
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Address");
          break;
        case 'v': /* vlan-name & vlan-num */
          switch(*(p->name + 6))
	  {
	      case 'a': /* vlan-name */
                 if(strcmp(p->name, "vlan-name") == 0)
	         {
 	            xmlSetProp(cur, "vlan-name", tmp_content);
                    xmlFreeNode(p);
                 }
                 else
                    badNode(p,"Address");
                 break;
	      case 'u': /* vlan-num */
                 if(strcmp(p->name, "vlan-num") == 0)
	         {
 	            xmlSetProp(cur, "vlan-num", tmp_content);
                    xmlFreeNode(p);
                 }
                 else
                    badNode(p,"Address");
                 break;
	      default:
	 	 badNode(p,"Address");
          }
          break;
        case 'a': /* address */
          if(strcmp(p->name, "address") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Address");
          break;
        case 'n': /* netmask */
          if(strcmp(p->name, "netmask") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Address");
          break;
        default:
          badNode(p,"Address");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);

}

/*==================================================================
 * Name: newUser
 *
 * Declaration: xmlNodePtr newUser(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident   : optional
 *   xmlNodePtr category: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr UserId: one or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created User node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newUser(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"User");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newUser: xmlNewNode(NULL,\"User\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"User");
          break;
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }          
          else
            badNode(p,"User");
          break;
        case 'U': /* UserId */
	  if(strcmp(p->name, "UserId") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"User");
          break;
        default:
          badNode(p,"User");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newUserId
 *
 * Declaration: xmlNodePtr newUserId(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident: optional
 *   xmlNodePtr type : optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr name  : zero or one
 *   xmlNodePtr number: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created User node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newUserId(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"UserId");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newUserId: xmlNewNode(NULL,\"UserId\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"UserId");
          break;
        case 't': /* type */
          if(strcmp(p->name, "type") == 0)
	  {
 	     xmlSetProp(cur, "type", tmp_content);
             xmlFreeNode(p);
          }          
          else
            badNode(p,"UserId");
          break;
        case 'n':
          switch(*(p->name + 1)) 
          {
             case 'a': /* name */
	       if(strcmp(p->name, "name") == 0)
	         xmlAddChild(cur, p);
               else
	         badNode(p,"UserId");
               break;
             case 'u': /* number */
	       if(strcmp(p->name, "number") == 0)
	         xmlAddChild(cur, p);
               else
	         badNode(p,"UserId");
               break;
             default:
               badNode(p,"UserId");
	  }
          break;
        default:
          badNode(p,"UserId");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newProcess
 *
 * Declaration: xmlNodePtr newProcess(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr name: exactly one
 *   xmlNodePtr pid : zero or one
 *   xmlNodePtr path: zero or one
 *   xmlNodePtr arg : zero or more
 *   xmlNodePtr env : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Process node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newProcess(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Process");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newProcess: xmlNewNode(NULL,\"Process\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Process");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Process");
          break;
        case 'p':
          switch(*(p->name + 1))
	  {
  	     case 'i': /* pid */
	       if(strcmp(p->name, "pid") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Process");
	       break;
	     case 'a': /* path */
	       if(strcmp(p->name, "path") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Process");
	       break;
             default:
               badNode(p,"Process");
	  }
          break;
        case 'a': /* arg */
          if(strcmp(p->name, "arg") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Process");
          break;
        case 'e': /* env */
          if(strcmp(p->name, "env") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Process");
          break;
        default:
          badNode(p,"Process");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newService
 *
 * Declaration: xmlNodePtr newService(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr name       : zero or one
 *   xmlNodePtr port       : zero or one
 *   xmlNodePtr portlist   : zero or one
 *   xmlNodePtr protocol   : zero or one
 *   xmlNodePtr SNMPService: zero or one
 *   xmlNodePtr WebService : zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Service node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newService(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Service");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newService: xmlNewNode(NULL,\"Service\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Service");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Service");
          break;
        case 'p':
          switch(*(p->name + 1))
	  {
  	     case 'r': /* protocol */
	       if(strcmp(p->name, "protocol") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Service");
	       break;
	     case 'o': /* portlist & port */
	       if((strcmp(p->name, "portlist") == 0) ||
                  (strcmp(p->name, "port") == 0))
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Service");
	       break;
             default:
               badNode(p,"Service");
	  }
          break;
        case 'S': /* SNMPService */
          if(strcmp(p->name, "SNMPService") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Service");
          break;
        case 'W': /* WebService */
          if(strcmp(p->name, "WebService") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Service");
          break;
        default:
          badNode(p,"Service");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newWebService
 *
 * Declaration: xmlNodePtr newWebService(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr url        : exactly one
 *   xmlNodePtr cgi        : zero or one
 *   xmlNodePtr http-method: zero or one
 *   xmlNodePtr arg        : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created WebService node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newWebService(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"WebService");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newWebService: xmlNewNode(NULL,\"WebService\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'u': /* url */
          if(strcmp(p->name, "url") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"WebService");
          break;
        case 'c': /* cgi */
	  if(strcmp(p->name, "cgi") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"WebService");
          break;
        case 'h': /* http-method */
	  if(strcmp(p->name, "http-method") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"WebService");
          break;
        case 'a': /* arg */
          if(strcmp(p->name, "arg") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"WebService");
          break;
        default:
          badNode(p,"WebService");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newSNMPService
 *
 * Declaration: xmlNodePtr newSNMPService(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr oid      : zero or one
 *   xmlNodePtr community: zero or one
 *   xmlNodePtr command  : zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly create SNMPService node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newSNMPService(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"SNMPService");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newSNMPService: xmlNewNode(NULL,\"SNMPService\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'o': /* oid */
          if(strcmp(p->name, "oid") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"SNMPService");
          break;
        case 'c':
          switch(*(p->name + 4))
	  {
  	     case 'u': /* community */
	       if(strcmp(p->name, "community") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"SNMPService");
	       break;
	     case 'a': /* command */
	       if(strcmp(p->name, "command") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"SNMPService");
	       break;
             default:
               badNode(p,"SNMPService");
	  }
          break;
        default:
          badNode(p,"SNMPService");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newFileList
 *
 * Declaration: xmlNodePtr newFileList(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr File: one or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly create FileList node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newFileList(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"FileList");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newFileList: xmlNewNode(NULL,\"FileList\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'F': /* File */
          if(strcmp(p->name, "File") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"FileList");
          break;
        default:
          badNode(p,"FileList");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newFile
 *
 * Declaration: xmlNodePtr newFile(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr ident   : optional
 *   xmlNodePtr category: required
 *   xmlNodePtr fstype  : required
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr name       : exactly one
 *   xmlNodePtr path       : exactly one
 *   xmlNodePtr create-time: Zero or one
 *   xmlNodePtr modify-time: Zero or one
 *   xmlNodePtr access-time: Zero or one
 *   xmlNodePtr disk-size  : Zero or one
 *   xmlNodePtr FileAccess : Zero or one
 *   xmlNodePtr Linkage    : Zero or one
 *   xmlNodePtr Inode      : Zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created File node.
 *
 * Example: similar to newAlert.
 *
 * NOTE: As the specification states, create-time, modify-time, and
 *       access-time are DATETIME formatted values.  These can be created
 *       using currentDatetime() or both newSimpleElement() and
 *       timevalToDatetime().  See the example1.c file for an example.
 *===================================================================*/

xmlNodePtr newFile(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"File");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newFile: xmlNewNode(NULL,\"File\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"File");
          break;
        case 'c':
          switch(*(p->name + 1))
	  {
  	     case 'a': /* category */
	       if(strcmp(p->name, "category") == 0)
		 {
		   xmlSetProp(cur, "category", tmp_content);
		   xmlFreeNode(p);
		 }
	       else
		 badNode(p,"File");
	       break;
	     case 'r': /* create-time */
	       if(strcmp(p->name, "create-time") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"File");
	       break;
             default:
               badNode(p,"File");
	  }
          break;
        case 'f': /* fstype */
          if(strcmp(p->name, "fstype") == 0)
	  {
 	     xmlSetProp(cur, "fstype", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"File");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"File");
          break;
        case 'p': /* path */
          if(strcmp(p->name, "path") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"File");
          break;
        case 'm': /* modify-time */
          if(strcmp(p->name, "modify-time") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"File");
          break;
        case 'a': /* access-time */
          if(strcmp(p->name, "access-time") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"File");
          break;
        case 'd': /* disk-size */
          if(strcmp(p->name, "disk-size") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"File");
          break;
        case 'F': /* FileAccess */
          if(strcmp(p->name, "FileAccess") == 0)
            xmlAddChild(cur,p);
          else
            badNode(p,"File");
          break;
        case 'L': /* Linkage */
          if(strcmp(p->name, "Linkage") == 0)
            xmlAddChild(cur,p);
          else
            badNode(p,"File");
          break;
        case 'I': /* Inode */
          if(strcmp(p->name, "Inode") == 0)
            xmlAddChild(cur,p);
          else
            badNode(p,"File");
          break;
        default:
          badNode(p,"File");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newFileAccess
 *
 * Declaration: xmlNodePtr newFileAccess(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr UserId    : Exactly one
 *   xmlNodePtr permission: one or more
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created FileAccess node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newFileAccess(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"FileAccess");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newFileAccess: xmlNewNode(NULL,\"FileAccess\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'U': /* UserID */
          if(strcmp(p->name, "UserId") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"FileAccess");
          break;
        case 'p': /* permission */
          if(strcmp(p->name, "permission") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"FileAccess");
          break;
        default:
          badNode(p,"FileAccess");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newLinkage
 *
 * Declaration: xmlNodePtr newLinkage(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr category: optional
 *
 *   SUB-ELEMENTS
 *   xmlNodePtr name: Exactly one
 *   xmlNodePtr path: Exactly one
 *   xmlNodePtr File: Exactly one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Linkage node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newLinkage(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Linkage");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newLinkage: xmlNewNode(NULL,\"Linkage\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"File");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Linkage");
          break;
        case 'p': /* path */
          if(strcmp(p->name, "path") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Linkage");
          break;
        case 'F': /* File */
          if(strcmp(p->name, "File") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Linkage");
          break;
        default:
          badNode(p,"Linkage");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newInode
 *
 * Declaration: xmlNodePtr newInode(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   SUB-ELEMENTS
 *   xmlNodePtr change-time   : zero or one
 *   xmlNodePtr number        : zero or one
 *   xmlNodePtr major-device  : zero or one
 *   xmlNodePtr minor-device  : zero or one
 *   xmlNodePtr c-major-device: zero or one
 *   xmlNodePtr c-minor-device: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Inode node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newInode(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Inode");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newInode: xmlNewNode(NULL,\"Inode\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'c':
          switch(*(p->name + 1))
	  {
  	     case 'h': /* change-time */
	       if(strcmp(p->name, "change-time") == 0)
		 xmlAddChild(cur, p);
 	       else
		 badNode(p,"Inode");		 
               break;
	     case '-': /* c-major-device or c-minor-device */
               switch(*(p->name + 3))
	       {
  	          case 'a': /* c-major-device */
	            if(strcmp(p->name, "c-major-device") == 0)
 	               xmlAddChild(cur, p);
	            else
		       badNode(p,"Inode");
	            break;
  	          case 'i': /* c-minor-device */
	            if(strcmp(p->name, "c-minor-device") == 0)
 	               xmlAddChild(cur, p);
	            else
		       badNode(p,"Inode");
	            break;
                  default:
                    badNode(p,"Inode");
               }
               break;
             default:
               badNode(p,"Inode"); 
          }
 	  break;
        case 'n': /* number */
          if(strcmp(p->name, "number") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Inode");
          break;
        case 'm':
          switch(*(p->name + 1))
	  {
  	     case 'a': /* major-device */
	       if(strcmp(p->name, "major-device") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Inode");
	       break;
	     case 'i': /* minor-device */
	       if(strcmp(p->name, "minor-device") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Inode");
	       break;
             default:
               badNode(p,"Inode");
	  }
          break;
        default:
          badNode(p,"Inode");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*********************************************************************
 *********************************************************************
 *****								 *****
 *****			The Assessment Classes			 *****
 *****								 *****
 *********************************************************************
 *********************************************************************/

/*==================================================================
 * Name: newImpact
 *
 * Declaration: xmlNodePtr newImpact(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr severity  : optional
 *   xmlNodePtr completion: optional
 *   xmlNodePtr type      : optional
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Impact node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newImpact(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Impact");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newImpact: xmlNewNode(NULL,\"Impact\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 's': /* severity */
          if(strcmp(p->name, "severity") == 0)
	  {
	     xmlSetProp(cur, "severity", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Impact");
          break;
	case 'c': /* completion */
	  if(strcmp(p->name, "completion") == 0)
	  {
	     xmlSetProp(cur, "completion", tmp_content);
             xmlFreeNode(p);
          }
	  else
	    badNode(p,"Impact");
	  break;
  	case 't': /* type */
          if(strcmp(p->name, "type") == 0)
	  {
	     xmlSetProp(cur, "type", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Impact");
          break;
        default:
          badNode(p,"Impact");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAction
 *
 * Declaration: xmlNodePtr newAction(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr category: required
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Action node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newAction(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Action");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAction: xmlNewNode(NULL,\"Action\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
	case 'c': /* category */
	  if(strcmp(p->name, "category") == 0)
	  {
	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }
	  else
	    badNode(p,"Action");
	  break;
        case 'v': /* value */
          if(strcmp(p->name, "value") == 0)
	  {
  	     xmlNodeSetContent(cur, tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Action");
          break;
        default:
          badNode(p,"Action");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newConfidence
 *
 * Declaration: xmlNodePtr newConfidence(xmlNodePtr, ...);
 *
 * Possible parameters: 
 * 
 *   ATTRIBUTES
 *   xmlNodePtr rating: required
 *
 * !Remember!: the last argument must be NULL.
 *
 * Returns: an xmlNodePtr to the newly created Confidence node.
 *
 * Example: similar to newAlert.
 *===================================================================*/

xmlNodePtr newConfidence(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Confidence");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newConfidence: xmlNewNode(NULL,\"Confidence\") failed\n");
    return NULL;
  }

  if(first == NULL) /* do we have at least one parameter? */
    return cur;

  /* cycle through the arguments */
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'r': /* rating */
          if(strcmp(p->name, "rating") == 0)
	  {
	     xmlSetProp(cur, "rating", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Confidence");
          break;
        case 'v': /* value */
          if(strcmp(p->name, "value") == 0)
	  {
  	     xmlNodeSetContent(cur, tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Confidence");
          break;
        default:
          badNode(p,"Confidence");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*********************************************************************
 *********************************************************************
 *****								 *****
 *****			Utility Functions			 *****
 *****								 *****
 *********************************************************************
 *********************************************************************/

/*==================================================================
 * Name: newSimpleElement
 *
 * Parameters 
 *  char *name : the parameter name
 *  char *value: the value of the parameter
 *===================================================================*/

xmlNodePtr newSimpleElement(char *name, char *value)
{
  xmlNodePtr cur;
  char *e_value;

  if((name == NULL) || (value == NULL))
    return NULL;

  cur = xmlNewNode(NULL,name);

  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newSimpleElement: xmlNewNode() failed\n");
    return NULL;
  }

  e_value = xmlEncodeEntitiesReentrant(doc,value); /* escape special chars */
  xmlNodeSetContent(cur,e_value);
  xmlFree(e_value);

  return(cur); 
}

/*==================================================================
 * Name: newAttribute
 *
 * Parameters 
 *  char *name : the parameter name
 *  char *value: the value of the parameter
 *
 * Note: This is the same function as newSimpleElement.  It was added to
 *       avoid confusion in building xml documents by following the
 *       nomenclature illustrated in the idmef-xml spec.  You may use
 *       newSimpleElement for adding new xml attributes if you so desire.
 *===================================================================*/

xmlNodePtr newAttribute(char *name, char *value)
{
  xmlNodePtr cur;
  char *e_value;
  
  if((name == NULL) || (value == NULL))
    return NULL;

  cur = xmlNewNode(NULL,name);

  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAttribute: xmlNewNode() failed\n");
    return NULL;
  }

  e_value = xmlEncodeEntitiesReentrant(doc,value);
  xmlNodeSetContent(cur,e_value);
  xmlFree(e_value);

  return(cur); 
}

/***********************************************************************
 *
 * These are functions that perform different operations on XML elements
 * that have been created with the above "new" functions.
 *
 ***********************************************************************/

/*===========================================================================
 * Name: addElement
 * Purpose: Add a child element to a parent.
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - child  : an xmlNodePtr to the child
 *
 * Returns: A pointer to the child, or NULL if error
 *
 * Note: We're not sure what type of error checking should occur here.  There's
 *       not really a way to ensure that the child should exist under the
 *       parent without implementing part of the DTD.  My initial thought is
 *       to let the DTD validation catch any errors at run time.
 *==========================================================================*/

xmlNodePtr addElement(xmlNodePtr parent, xmlNodePtr child)
{
  return xmlAddChild(parent, child);
}

/*===========================================================================
 * Name: addElements
 * Purpose: Add a child elements to a parent.
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - child  : an xmlNodePtr to a child
 *           - ...    : a variable size list of xmlNodePtrs to children
 *
 * Returns: the number of children added
 *
 * NOTE: The last parameter MUST be NULL.  Otherwise, the function will not
 *       know when to stop reading in the variable length parameter list, and
 *       bad things will happen.
 *==========================================================================*/

int addElements(xmlNodePtr parent, xmlNodePtr child, ...)
{
  xmlNodePtr p;
  va_list ap;
  int count = 0;

  if((parent == NULL) || (child == NULL)) /* do we have at least two parameters? */
    return 0;

  /* cycle through the arguments */
  va_start(ap, child);

  for(p = child; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name != NULL)
     {
        xmlAddChild(parent, p);
        count++;
     }
  }
  va_end(ap);

  return(count);
}

/*===========================================================================
 * Name: setAttribute
 * Purpose: Set attributes to an element.
 * Parameters:
 *           - element : an xmlNodePtr to the element
 *           - attrib  : an xmlNodePtr to the attribute
 *
 * Returns: A pointer to the element, or NULL if error
 *==========================================================================*/

xmlNodePtr setAttribute(xmlNodePtr element, xmlNodePtr attrib)
{
  char *propName = NULL;
  xmlChar *tmp_content = NULL;
  propName = xmlStrdup(attrib->name);

  tmp_content = xmlNodeGetContent(attrib);

  if(xmlSetProp(element, propName, tmp_content))
  {
     xmlFreeNode(attrib);
     if(tmp_content != NULL)
        xmlFree(tmp_content);
     if(propName != NULL)
       xmlFree(propName);
     return element;
  }

  if(tmp_content != NULL)
     xmlFree(tmp_content);

  if(propName != NULL)
     xmlFree(propName);

  xmlFreeNode(attrib);
  return NULL;
}

/*===========================================================================
 * Name: setAttributes
 * Purpose: Set attributes to a element.
 * Parameters:
 *           - element : an xmlNodePtr to the element
 *           - attrib  : an xmlNodePtr to an attribute
 *           - ...     : a variable size list of xmlNodePtrs to attributes
 *
 * Returns: the number of attributes set
 *
 * NOTE: The last parameter MUST be NULL.  Otherwise, the function will not
 *       know when to stop reading in the variable length parameter list, and
 *       bad things will happen.
 *==========================================================================*/

int setAttributes(xmlNodePtr element, xmlNodePtr attrib, ...)
{
  xmlNodePtr p;
  va_list ap;
  char *propName = NULL;
  xmlChar *tmp_content = NULL;
  int count = 0;


  /* do we have at least two parameters? */
  if((element == NULL) || (attrib == NULL))
    return 0;

  /* cycle through the arguments */
  va_start(ap, attrib);

  for(p = attrib; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     tmp_content = xmlNodeGetContent(p);

     if(p->name != NULL)
     {
        propName = xmlStrdup(p->name);

        xmlSetProp(element, propName, tmp_content);
        count++;
     }
     xmlFreeNode(p);
 
     /* these values were copied in xmlSetProp, so free 'em */
     if(tmp_content != NULL)
        xmlFree(tmp_content);
     if(propName != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(count);
}

/*==========================================================================*
 * Name: getElement()
 * Purpose: Retrieve a pointer to the first child element directly under a
 *          parent element whose name matches the provided string.
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - name   : a string representing the child's name
 *
 * Returns: A xmlNodePtr containing the element.
 *==========================================================================*/

xmlNodePtr getElement(xmlNodePtr parent, char *name)
{
  xmlNodePtr p;

  /* cycle through the children */
  for(p = parent->children; p != NULL; p = p->next)
  {
     if(strcmp(p->name, name) == 0)
        return p;
  }  
  return NULL;
}

/*==========================================================================*
 * Name: hasElement()
 * Purpose: Determines whether a parent element has a specified child element
 *          directly below it.  This is to say it only searches for first
 *          generation children.
 *
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - name   : a string representing the child's name
 *
 * Returns: 1 for yes, 0 for no
 *==========================================================================*/

int hasElement(xmlNodePtr parent, char *name)
{
  xmlNodePtr p;

  /* cycle through the children */
  for(p = parent->children; p != NULL; p = p->next)
  {
     if(strcmp(p->name, name) == 0)
        return 1;
  }  
  return 0;
}

 /***********************************************************************
  *
  * These are functions that perform different operations on XML documents.
  *
  ***********************************************************************/

/*==========================================================================*
 * Name: globalsInit()
 *
 * Purpose: Initialize global variables, and those structures we can reuse
 *          for each IDMEF Message.
 *
 * Parameters:
 *           - xml_version : The current version of XML
 *
 * Returns: 1 for success, and 0 otherwise
 *
 * NOTE: You MUST call this function before you start to create IDMEF messages
 *==========================================================================*/
int globalsInit(char *dtd_path)
{
  doc = NULL;
  g_output = NULL;

  xmlInitMemory();

  g_dtd_path = dtd_path;

  /* Initialize the validity context */

  vctxt = malloc(sizeof(xmlValidCtxt));

  if(vctxt == NULL)
  {
     fprintf(stderr,"idmefxml.c : globalsInit() : malloc fail!\n");
     return 0;
  }
  vctxt->userData = (void *) stderr;
  vctxt->error = (xmlValidityErrorFunc) fprintf;
  vctxt->warning = (xmlValidityWarningFunc) fprintf;

  /* Initialize the dtd */

  dtd = xmlParseDTD(NULL,dtd_path);

  if(dtd == NULL)
  {
    fprintf(stderr,"idmefxml.c: createDoc(): problem parsing the DTD\n");
    return 0;
  }
  dtd->name = xmlStrdup((xmlChar*)DTD_NAME);

  return 1;
}

/*==========================================================================*
 * Name: createCurrentDoc()
 *
 * Purpose: Creates the a new xmlDoc and sets it as the current (doc).
 *
 * Parameters:
 *           - xml_version : The current version of XML
 *
 * Returns: 1 for success, and 0 otherwise
 *==========================================================================*/
int createCurrentDoc(const char *xml_version)
{
  xmlDocPtr thisDoc;
  thisDoc = createDoc(xml_version);
  
  if(setCurrentDoc(thisDoc))
     return 1;
  return 0;
}

/*==========================================================================*
 * Name: createDoc()
 * Purpose: Creates a new xmlDoc, initializes the root element, and links
 *          to the DTD Internal Subset.
 *
 * Parameters:
 *           - xml_version : The current version of XML
 *
 * Returns: an xmlDocPtr to the newly created doc, or NULL for error.
 *==========================================================================*/

xmlDocPtr createDoc(const char *xml_version)
{
  xmlDocPtr newDoc;
  xmlDtdPtr intSubset;
  xmlNodePtr root, oldRoot;
  
  /* Initialize the document and root element */

  newDoc = xmlNewDoc(xml_version);
  
  if(newDoc == NULL)
  {
    fprintf(stderr,"idmefxml.c: createDoc(): problem creating new xml doc\n");
    return 0;
  }

  /* initialize the Internal Subset */
                  
  intSubset = xmlCreateIntSubset(newDoc,
                  (xmlChar*)DTD_NAME,
                  (xmlChar*)DTD_EXTERNAL_ID,
                  (xmlChar*)g_dtd_path);

  /* Create the root node and attach it */

  root = xmlNewNode(NULL,"IDMEF-Message");

  if(root == NULL)
  {
    fprintf(stderr,"idmefxml.c: newIDMEFDoc(): problem creating the root node\n");
    return NULL;
  }

  oldRoot = xmlDocSetRootElement(newDoc,root); 

  if(oldRoot != NULL)
    xmlFreeNode(oldRoot);

  return newDoc;
}

/*==========================================================================*
 * Name: setCurrentDoc()
 * 
 * Purpose : Sets the current doc and free's the old doc if need be.
 *
 * Parameters:
 *         - newDoc : An xmlDocPtr to the doc to be set as the current
 *
 * Returns: 1 if success, 0 otherwise
 *==========================================================================*/
int setCurrentDoc(xmlDocPtr newDoc)
{
   if(newDoc == NULL)
   {
      fprintf(stderr,"idmefxml.c : setCurrentDoc() - NULL parameter newDoc\n");
      return 0;
   }
   
   if(doc != NULL)      /* was there a current doc already? */
   { 
      clearCurrentDoc();
   }

   doc = newDoc;

   return 1;
}

/*==========================================================================*
 * Name: getCurrentDoc()
 * Purpose: gets the current xmlDoc (the global "doc").
 *
 * Returns: an xmlDocPtr to the current doc.
 *==========================================================================*/
xmlDocPtr getCurrentDoc()
{
  return doc;
}

/*==========================================================================*
 * Name: clearCurrentDoc()
 *
 * Purpose: Free the current xmlDoc, including all it's structures.
 *
 * NOTE: Originally, this function did not free the DTD.  This has been changed
 *       due to the fact that without the DTD, there is not a DTD document
 *       type declaration, which makes the whole document invalid  (as defined
 *       in section 4.1.4 of draft-ietf-idwg-iemef-xml-01.txt).
 *==========================================================================*/
void clearCurrentDoc()
{
    if(doc == NULL)
      return;

    xmlFreeDoc(doc);    

    doc = NULL;
    xmlCleanupParser();
}

/*==========================================================================*
 * Name: copyCurrentDoc()
 *
 * Purpose: Create a copy of the current xmlDoc.  This copies everything,
 *          including the internal subset.
 *
 * Returns:
 *        - docCopy : A copy of the current xml doc
 *
 * NOTE: Originally, this function did not copy the DTD.  This has been changed
 *       due to the fact that without the DTD, there is not a DTD document
 *       type declaration, which makes the whole document invalid  (as defined
 *       in section 4.1.4 of draft-ietf-idwg-iemef-xml-01.txt).
 *==========================================================================*/
xmlDocPtr copyCurrentDoc()
{
  xmlDocPtr docCopy;

  docCopy = xmlCopyDoc(doc,1);  /* recursively copy */

  return docCopy;
}

/*==========================================================================*
 * Name: resetCurrentDoc()
 *
 * Purpose: Reset the current document by clearing it and re-initializing
 *          it.
 *
 * Returns: 1 for success, and 0 otherwise
 *==========================================================================*/
int resetCurrentDoc()
{
  xmlChar *theVersion;

  theVersion = xmlStrdup(doc->version);

  clearCurrentDoc();

  if(createCurrentDoc(theVersion))
  {
     if(theVersion != NULL)
        xmlFree(theVersion);
     return 1;
  }

  if(theVersion != NULL)
     xmlFree(theVersion);

  return 0;  
}

/*==========================================================================*
 * Name: validateCurrentDoc()
 *
 * Purpose: Validates the current document
 *
 * Returns: 1 for success, and 0 otherwise
 *==========================================================================*/
int validateCurrentDoc()
{
  if(!xmlValidateDtd(vctxt,doc,dtd))  
    return 0;
  return 1;
}

/*==========================================================================*
 * Name: printCurrentMessage()
 *
 * Purpose: Prints the current message to the FILE * provided
 *
 * CHANGES: changed from xmlElemDump(f,doc,doc->children), because
 *          we were not printing the XML declaration, making it invalid.
 *
 *==========================================================================*/
void printCurrentMessage(FILE *f)
{
   if(g_output == NULL) /* haven't initialized the global output buffer yet? */
   {
     if(f == NULL) /* ok, then we better have an open file handle */
       {
	  fprintf(stderr,"printCurrentMessage: You must pass in an open FILE argument the first time this function is called.  This will initialize the file output buffer\n");
          return;
       }
     g_output = IDMEFxmlOutputBufferCreate(f, NULL); /* create the buffer.. */
   }

   IDMEFxmlDocContentDumpOutput(g_output, doc, NULL, 1); /* rinse... */
   xmlOutputBufferFlush(g_output);                       /* and flush */
}

/***********************************************************************
 *
 *               Miscellaneous time format functions.
 *
 ***********************************************************************/

/*==========================================================================*
 * Name: currentNtpTimestamp()
 *
 * Purpose: Creates a NTP timestamp with the current time
 *
 * Returns: The current NTP timestamp as a char *
 *
 *==========================================================================*/

xmlChar *currentNtpTimestamp()
{
   struct timeval currtv;

   gettimeofday(&currtv,0);
   return timevalToNtpTimestamp(&currtv);
}

/*==========================================================================*
 * Name: currentDatetime
 *
 * Purpose: Formats the current datetime as specified by the IDMEF XML
 *          specification version 3 (4.4.6). YYYY-MM-DDThh:mm:ss(+|-)hh:mm
 *
 * Declaration: char *currentDatetime();
 *
 * Returns: The formated date as a char *
 *
 * Example: 
 *          char *datetime = currentDatetime();
 *==========================================================================*/

xmlChar *currentDatetime()
{
   xmlChar *time_str;
   time_t curr_time;
   struct tm *loc_time;

   curr_time = time(NULL);
   loc_time = localtime(&curr_time);

   time_str = (xmlChar *) xmlMalloc(MAX_UTC_DATETIME_SIZE * sizeof(xmlChar));
   strftime(time_str,MAX_UTC_DATETIME_SIZE,"%Y-%m-%dT%H:%M:%SZ",loc_time);
 
   return time_str;
}

/*==========================================================================*
 * Name: timevalToNtpTimestamp
 *
 * Purpose: Creates a NTP timestamp with the provided timeval struct
 *
 * Declaration: char *convertNtpTimestamp(const time_t *);
 *
 * Returns: The provided NTP timestamp as a char *
 *
 * Example:
 *          struct timeval t;
 *          char *ntp_timestamp_string;
 *
 *          gettimeofday(&t,0);
 *          ntp_timestamp_string = timevalToNtpTimestamp(&t);
 *==========================================================================*/

xmlChar *timevalToNtpTimestamp(struct timeval *t)
{
   xmlChar *idmef_ntp_timestamp;
   long seconds, microseconds;

   seconds = t->tv_sec + 2208988800UL;
   microseconds = (4294967296 * t->tv_usec) / 1000000;

   idmef_ntp_timestamp = (xmlChar *) xmlMalloc(sizeof(xmlChar) + MAX_NTP_TIMESTAMP_SIZE); /* 0xNNNNNNNN.0xNNNNNNNN */
   sprintf(idmef_ntp_timestamp,"0x%lx.0x%lx", seconds, microseconds);

   return idmef_ntp_timestamp;
}

/*==========================================================================*
 * Name: timevalToDatetime *
 * Purpose: Formats the provided time as specified by the IDMEF XML
 *          specification version 3 (4.4.6). YYYY-MM-DDThh:mm:ssZ.  Time
 *          is represented as UTC.
 *
 * Declaration: char *convertTime(const time_t *);
 *
 * Returns: The formated time as a char *
 *
 * Example:
 *          time_t t;
 *          char *datetime_string;
 *
 *          t = time(NULL);
 *          datetime_string = timeToDatetime(&t);
 *
 * Note: This function outputs time in UTC, partially because we technically
 *       save space and bandwith, but more so because strftime's representation
 *       of "time-zone as hour offset from GMT", excludes the ':', which is
 *       required in the IDMEF spec.
 *==========================================================================*/

xmlChar *timevalToDatetime(struct timeval *t)
{
   xmlChar *date_str;
   struct tm *loc_time;

   loc_time = gmtime(&t->tv_sec);
   
   date_str = (xmlChar *) xmlMalloc(MAX_UTC_DATETIME_SIZE * sizeof(xmlChar));
   strftime(date_str,MAX_UTC_DATETIME_SIZE,"%Y-%m-%dT%H:%M:%SZ",loc_time);
 
   return date_str;
}

/***********************************************************************
 *
 *                   Other utility functions
 *
 ***********************************************************************/

/*==========================================================================*
 * Name: getStoredAlertID()
 *
 * Purpose: Opens the specified file, and attempts to extract the stored
 *          alert id.
 *
 * Returns: the alert id as an unsigned long integer, 1 if the file didn't
 *          contain an alert id, or 0 for error.
 *==========================================================================*/
ulong getStoredAlertID(char *f_name)
{
  FILE *f;
  char *idmef_alertid_string;
  ulong aid;

  /* Open the alert ID file and extract the stored alert ID if it exists*/
  f = fopen(f_name,"a+");
  if (!f) 
  {
     fprintf(stderr,"getStoredAlertID: IDMEF alert ID file open error (%s)\n",f_name);
     return 0;
  }

  /** Now seek back to the beginning... Nice little hack **/
  rewind(f);
    
  idmef_alertid_string = (char *) calloc(MAX_ALERTID_BUFFER_SIZE,sizeof(char));

  if(fgets(idmef_alertid_string, MAX_ALERTID_BUFFER_SIZE + 1, f) == NULL)
  {
    /*     fprintf(stderr,"getStoredAlertID: Stored alert ID not found in %s, continuing with alert ID = 1\n", f_name); */
     aid = 1;
  }
  else
     aid = (ulong) atol(idmef_alertid_string);
  
  return aid;

  fclose(f); /* done for now */
}

/*==========================================================================*
 * Name: saveAlertID()
 *
 * Purpose: Opens the specified file, and attempts to extract the stored
 *          alert id.
 *
 * Returns: 1 for success or 0 for error.
 *==========================================================================*/
int saveAlertID(ulong aid, char *f_name)
{
   char *temp;
   FILE *f;
 
   temp = (char *) calloc(MAX_ALERTID_BUFFER_SIZE,sizeof(char));
   temp = ulongToString(aid);

   f = fopen(f_name,"w");
   if (!f) 
   {
      fprintf(stderr,"saveAlertID: IDMEF alert ID file open error (%s)\n",f_name);
      return 0;
   }

   if(fputs(ulongToString(aid), f) < 0)
   {
     fprintf(stderr,"saveAlertID: Error writting alert id (%lu) to %s\n", aid, f_name);
     return 0;
   }
   fputc('\n',f);
   fflush(f);
   fclose(f);

   return 1;
}

/*==========================================================================*
 * Name: badNode()
 * Purpose: Prints a message to stderr with the invalid node name, and the
 *          parent it was attempted to be linked to.
 *
 * Parameters:
 *           - theNode : the invalid node
 *           - parent  : the would-be parent node's name
 *==========================================================================*/
void badNode(xmlNodePtr theNode, char *parent)
{
   fprintf(stderr, "idmefxml.c: Invalid xmlNode for parent %s\n", parent);
   fprintf(stderr, "idmefxml.c: Invalid xmlNode \"%s\" for %s\n",
          theNode->name, parent);
   xmlFreeNode(theNode);
   return;
}

/*==========================================================================*
 * Name: intToString()
 * Purpose: Converts a signed integer value to it's ASCII representation.
 *
 * Parameters:
 *           - int_val : the value to be converted
 *
 * Returns: The ASCII representation
 *==========================================================================*/
char * intToString(int int_val)
{
  int ret, tmp;
  char *string_val;

  if(int_val > 0)
    tmp = floor(log10(int_val)) + 2;
  else
     if(int_val < 0)
        tmp = floor(log10(-1 * int_val)) + 3;
     else
        tmp = 2;

  string_val = calloc(tmp,sizeof(char));
  if(string_val == NULL)
     return NULL;
  ret = sprintf(string_val,"%d",int_val);

  if(ret > -1 && ret <= tmp)
     return string_val;
  
  return NULL;
}

/*==========================================================================*
 * Name: ulongToString()
 * Purpose: Converts an unsigned value to it's ASCII representation.
 * 
 * Parameters:
 *           - ulong_val : the value to be converted
 *
 * Returns: The ASCII representation
 *
 * Note: This function was custom made for converting ulong alertID's to
 *       strings.
 *==========================================================================*/
char * ulongToString(ulong ulong_val)
{
  int ret;
  ulong tmp;
  char *string_val;

  if(ulong_val == 0)
    tmp = 2;
  else
    tmp = floor(log10(ulong_val)) + 2;
 
  string_val = calloc(tmp,sizeof(char));
  if(string_val == NULL)
     return NULL;
  ret = sprintf(string_val,"%lu",ulong_val);

  if(ret > -1 && ret <= tmp)
     return string_val;

  return NULL;
}

/*=========================================================================*
 * Name: IDMEFxmlOutputBufferCreate:
 * Purpose: Create a buffered  output for the progressive saving of a file
 *          If filename is "-' then we use stdout as the output.
 *          Automatic support for ZLIB/Compress compressed document is provided
 *          by default if found at compile-time.
 *
 * Parameters:
 *             @URI:  a C string containing the URI or filename
 *             @encoder:  the encoding converter or NULL
 *             @compression:  the compression ration (0 none, 9 max).
 *
 * Returns: the new output or NULL
 * Note: This function is another strain of the xmlOutputBufferCreate functions
 *       from libxml.  It's also happens to be used in an I/0 example from
 *       www.xmlsoft.org, and happens to fit our needs nicely.
 *=========================================================================*/
xmlOutputBufferPtr IDMEFxmlOutputBufferCreate(FILE *file, xmlCharEncodingHandlerPtr encoder) 
{
           xmlOutputBufferPtr ret;
           
           xmlRegisterDefaultOutputCallbacks();

           if (file == NULL) return(NULL);

           ret = xmlAllocOutputBuffer(encoder);
           if (ret != NULL) {
               ret->context = file;
               ret->writecallback = IDMEFxmlFileWrite;
               ret->closecallback = NULL;  /* No close callback */
           }
           return(ret); 
} 

/*==========================================================================*
 * Name: IDMEFxmlFileWrite
 * Purpose: Writes @len bytes from @buffer to the I/O channel
 * 
 * Parameters:
 *            @context:  the I/O context
 *            @buffer:  where to drop data
 *            @len:  number of bytes to write
 *
 * Returns: The number of bytes written
 *
 * Note: This function is an import of a static function called xmlFileWrite
 *       from libxml.
 *==========================================================================*/

int IDMEFxmlFileWrite (void * context, const char * buffer, int len) {
    return(fwrite(&buffer[0], 1,  len, (FILE *) context));
}

/*=========================================================================*
 * Name: IDMEFxmlDocContentDumpOutput:
 * Purpose: Dumps an xml Document to an xmlOutputBufferPtr
 *
 * Parameters:
 *             @buf:  the XML buffer output
 *             @cur:  the document
 *             @encoding:  an optional encoding string
 *             @format:  should formatting spaces been added
 *
 * Note: This function is an import of a static function called
 *       xmlDocContentDumpOutput from libxml.
 *==========================================================================*/
void IDMEFxmlDocContentDumpOutput(xmlOutputBufferPtr buf, xmlDocPtr cur,
                             const char *encoding, int format) 
{
    xmlOutputBufferWriteString(buf, "<?xml version=");
    if (cur->version != NULL)
        xmlBufferWriteQuotedString(buf->buffer, cur->version);
    else
        xmlOutputBufferWriteString(buf, "\"1.0\"");
    if (encoding == NULL) {
        if (cur->encoding != NULL)
            encoding = (const char *) cur->encoding;
        else if (cur->charset != XML_CHAR_ENCODING_UTF8)
            encoding = xmlGetCharEncodingName((xmlCharEncoding) cur->charset);
    }
    if (encoding != NULL) {
        xmlOutputBufferWriteString(buf, " encoding=");
        xmlBufferWriteQuotedString(buf->buffer, (xmlChar *) encoding);
    }
    switch (cur->standalone) {
        case 0:
            xmlOutputBufferWriteString(buf, " standalone=\"no\"");
            break;
        case 1:
            xmlOutputBufferWriteString(buf, " standalone=\"yes\"");
            break;
    }
    xmlOutputBufferWriteString(buf, "?>\n");
    if (cur->children != NULL) {
        xmlNodePtr child = cur->children;

        while (child != NULL) {
            xmlNodeDumpOutput(buf, cur, child, 0, format, encoding);
            xmlOutputBufferWriteString(buf, "\n");
            child = child->next;
        }
    }
}
