/*
idmefxml_gen.c: libidmef message generation functions
Author: Adam Migus NAI Labs (amigus@nai.com)

Copyright (c) 2001 Networks Associates Technology, Inc.
All rights reserved

This library is released under the GNU GPL and BSD software licenses.
You may choose to use one or the other, BUT NOT BOTH.  The GNU GPL
license is located in the file named COPYING.  The BSD license is located
in the file named COPYING.BSD.  Please contact us if there are any
questions.

This code creates messages that adhere to the specification put forth by
the IETF IDWG in the standards document draft-ietf-idwg-idmef-xml-06.txt.

$Id: idmefxml_gen.c,v 1.5 2002/02/20 23:54:28 amigus Exp $
*/

#include <stdio.h>
#include <string.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <stdlib.h>
#include <netdb.h>

#include "idmefxml.h"
#include "idmefxml_parse.h"
#include "idmefxml_gen.h"

#ifdef IDMEF_DEBUG
#define DEBUG 1
#else
#define DEBUG 0
#endif /* IDMEF_DEBUG */

char *text = 0;
int text_length = 0;

/*
** Name: build_XXX where XXX is an element type defined in the
** IDMEF message exchange format specification.  Types are in
** lowercase in all cases.
**
** Purpose: The build_XXX function builds an XML representation
** of the structure of type XXX.
** 
** Arguments: In all cases the only argument is the C structure
** of type XXX.
**
** Returns: An XML Node pointer where the root node is type IDMEF
** type XXX.
**
** Notes:
**
*/

xmlNodePtr
build_address(IDMEFaddress *address) {
	xmlNodePtr cur = 0;

	cur = newAddress(NULL);

	if(address->ident)
		setAttribute(cur, newAttribute("ident", address->ident));
	
	if(address->category != -1)
		setAttribute(cur, newAttribute("category",
		address_category_as_string(address->category)));

	if(address->vlan_name)
		setAttribute(cur, newAttribute("vlan-name",
			address->vlan_name));

	if(address->vlan_num) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", address->vlan_num);
		setAttribute(cur,
		newAttribute("vlan-num", buf));
	}

	if(address->address)
		addElement(cur,
		newSimpleElement("address", address->address));

	if(address->netmask)
		addElement(cur,
		newSimpleElement("netmask", address->netmask));

	return cur;
}

xmlNodePtr
build_userid(IDMEFuserid *userid) {
	xmlNodePtr cur = 0;

	cur = newUserId(NULL);

	if(userid->ident)
		setAttribute(cur, newAttribute("ident", userid->ident));

	if(userid->type != -1)
		setAttribute(cur, newAttribute("type",
		userid_type_as_string(userid->type)));

	if(userid->name)
		addElement(cur,
		newSimpleElement("name", userid->name));

	if(userid->number) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", userid->number);
		addElement(cur,
		newSimpleElement("number", buf));
	}

	return cur;
}

xmlNodePtr
build_user(IDMEFuser *user) {
	xmlNodePtr cur = 0;
	int i = 0;

	cur = newUser(NULL);

	if(user->ident)
		setAttribute(cur, newAttribute("ident", user->ident));

	if(user->category != -1)
		setAttribute(cur, newAttribute("category",
		user_category_as_string(user->category)));

	for(i = 0; i < user->nuserids; i++) {
		addElement(cur, build_userid(user->userids[i]));
	}

	return cur;
}

xmlNodePtr
build_SNMPservice(IDMEFSNMPservice *SNMPservice) {
	xmlNodePtr cur = 0;

	cur = newSNMPService(NULL);

	if(SNMPservice->oid)
		addElement(cur,
		newSimpleElement("oid", SNMPservice->oid));

	if(SNMPservice->community)
		addElement(cur,
		newSimpleElement("community", SNMPservice->community));

	if(SNMPservice->command)
		addElement(cur,
		newSimpleElement("command", SNMPservice->command));

	return cur;
}

xmlNodePtr
build_webservice(IDMEFwebservice *webservice) {
	xmlNodePtr cur = 0;
	int i = 0;

	cur = newWebService(NULL);

	if(webservice->url)
		addElement(cur,
		newSimpleElement("url", webservice->url));

	if(webservice->cgi)
		addElement(cur,
		newSimpleElement("cgi", webservice->cgi));

	if(webservice->http_method)
		addElement(cur,
		newSimpleElement("http-method", webservice->http_method));

	for(i = 0; i < webservice->narg; i++) {
		addElement(cur,
		newSimpleElement("arg", webservice->arg[i]));
	}

	return cur;
}

xmlNodePtr
build_service(IDMEFservice *service) {
	xmlNodePtr cur = 0;

	cur = newService(NULL);

	if(service->ident)
		setAttribute(cur, newAttribute("ident", service->ident));

	if(service->name)
		addElement(cur,
		newSimpleElement("name", service->name));


	if(service->port) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", service->port);
		addElement(cur,
		newSimpleElement("port", buf));
	}

	if(service->portlist)
		addElement(cur,
		newSimpleElement("portlist", service->portlist));

	if(service->protocol) {
		char buf[BUFSIZ];
		struct protoent *p = 0;

		memset(buf, '\0', sizeof(buf));

		if(!(p = getprotobynumber(service->protocol)))
			sprintf(buf, "%d", service->protocol);
		else
			sprintf(buf, "%s", p->p_name);
		addElement(cur,
		newSimpleElement("protocol", buf));
	}

	if(service->webservice)
		addElement(cur,
		build_webservice(service->webservice));

	if(service->snmpservice)
		addElement(cur,
		build_SNMPservice(service->snmpservice));

	return cur;
}

xmlNodePtr
build_process(IDMEFprocess *process) {
	xmlNodePtr cur = 0;
	int i;

	cur = newProcess(NULL);

	if(process->ident)
		setAttribute(cur, newAttribute("ident", process->ident));

	if(process->name)
		addElement(cur, 	
		newSimpleElement("name", process->name));

	if(process->pid) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", process->pid);
		addElement(cur,
		newSimpleElement("pid", buf));
	}

	if(process->path)
		addElement(cur, 	
		newSimpleElement("path", process->path));

	for(i = 0; i < process->narg; i++) {
		addElement(cur,
		newSimpleElement("arg", process->arg[i]));
	}

	for(i = 0; i < process->nenv; i++) {
		addElement(cur,
		newSimpleElement("env", process->env[i]));
	}

	return cur;
}

xmlNodePtr
build_node(IDMEFnode *node) {
	xmlNodePtr cur = 0;
	int i;

	cur = newNode(NULL);

	if(node->ident)
		setAttribute(cur, newAttribute("ident", node->ident));

	if(node->category != -1)
		setAttribute(cur, newAttribute("category",
		node_category_as_string(node->category)));

	if(node->location)
		addElement(cur,
		newSimpleElement("location", node->location));

	if(node->name)
		addElement(cur,
		newSimpleElement("name", node->name));

	for(i = 0; i < node->naddresses; i++) {
		addElement(cur, build_address(node->addresses[i]));
	}

	return cur;
}

xmlNodePtr
build_source(IDMEFsource *source) {
	xmlNodePtr cur = 0;

	cur = newSource(NULL);

	if(source->ident)
		setAttribute(cur, newAttribute("ident", source->ident));
	
	if(source->spoofed != -1)
		setAttribute(cur, newAttribute("spoofed",
		spoofed_as_string(source->spoofed)));

	if(source->interface)
		setAttribute(cur, newAttribute("interface",
			source->interface));
	
	if(source->node)
		addElement(cur, build_node(source->node));
		
	if(source->user)
		addElement(cur, build_user(source->user));

	if(source->process)
		addElement(cur, build_process(source->process));

	if(source->service)
		addElement(cur, build_service(source->service));

	return cur;
}

xmlNodePtr
build_fileaccess(IDMEFfileaccess *fileaccess) {
	xmlNodePtr cur = 0;
	int i;

	cur = newFileAccess(NULL);

	if(fileaccess->userid)
		addElement(cur, build_userid(fileaccess->userid));

	for(i = 0; i < fileaccess->npermissions; i++) {
		addElement(cur, newSimpleElement("permission",
		fileaccess->permissions[i]));
	}

	return cur;
}

xmlNodePtr
build_linkage(IDMEFlinkage *linkage) {
	xmlNodePtr cur = 0;

	cur = newLinkage(NULL);

	if(linkage->category != -1)
		setAttribute(cur, newAttribute("category",
		linkage_category_as_string(linkage->category)));

	if(linkage->name)
		addElement(cur, newSimpleElement("name", linkage->name));

	if(linkage->path)
		addElement(cur, newSimpleElement("path", linkage->path));

	if(linkage->file)
		addElement(cur, build_file(linkage->file));

	return cur;
}

xmlNodePtr
build_inode(IDMEFinode *inode) {
	xmlNodePtr cur = 0;

	cur = newInode(NULL);

	//if(inode->change_time)
	//	addElement(cur, newInodeChangeTime(&inode->change_time->tv));

	if(inode->number > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", inode->number);
		addElement(cur, newSimpleElement("number", buf));
	}

	if(inode->major_device > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", inode->major_device);
		addElement(cur, newSimpleElement("major-device", buf));
	}

	if(inode->minor_device > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", inode->minor_device);
		addElement(cur, newSimpleElement("minor-device", buf));
	}

	if(inode->c_major_device > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", inode->c_major_device);
		addElement(cur, newSimpleElement("c-major-device", buf));
	}

	if(inode->c_minor_device > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", inode->c_minor_device);
		addElement(cur, newSimpleElement("c-minor-device", buf));
	}

	return cur;
}

xmlNodePtr
build_file(IDMEFfile *file) {
	xmlNodePtr cur = 0;
	int i;

	cur = newFile(NULL);

	if(file->ident)
		setAttribute(cur, newAttribute("ident", file->ident));

	if(file->category != -1)
		setAttribute(cur, newAttribute("category",
		file_category_as_string(file->category)));

	if(file->fstype)
		setAttribute(cur, newAttribute("fstype", file->fstype));

	if(file->name)
		addElement(cur, newSimpleElement("name", file->name));

	if(file->path)
		addElement(cur, newSimpleElement("path", file->name));

	//if(file->create_time)
	//	addElement(cur, newFileCreateTime(&file->create_time->tv));

	//if(file->modify_time)
	//	addElement(cur, newFileModifyTime(&file->modify_time->tv));

	//if(file->access_time)
	//	addElement(cur, newFileAccessTime(&file->access_time->tv));

	if(file->data_size > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", file->data_size);
		addElement(cur, newSimpleElement("data-size", buf));
	}

	if(file->disk_size > -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", file->disk_size);
		addElement(cur, newSimpleElement("disk-size", buf));
	}

	for(i = 0; i < file->nfileaccesses; i++) {
		addElement(cur, build_fileaccess(file->fileaccesses[i]));
	}

	for(i = 0; i < file->nlinkages; i++) {
		addElement(cur, build_linkage(file->linkages[i]));
	}

	if(file->inode)
		addElement(cur, build_inode(file->inode));

	return cur;
}

xmlNodePtr
build_filelist(IDMEFfilelist *filelist) {
	xmlNodePtr cur = 0;
	int i;

	cur = newFileList(NULL);

	for(i = 0; i < filelist->nfiles; i++) {
		addElement(cur, build_file(filelist->files[i]));
	}

	return cur;
}

xmlNodePtr
build_target(IDMEFtarget *target) {
	xmlNodePtr cur = 0;

	cur = newTarget(NULL);

	if(target->ident)
		setAttribute(cur, newAttribute("ident", target->ident));
	
	if(target->decoy != -1)
		setAttribute(cur, newAttribute("decoy",
		decoy_as_string(target->decoy)));

	if(target->interface)
		setAttribute(cur, newAttribute("interface",
			target->interface));
	
	if(target->node)
		addElement(cur, build_node(target->node));
		
	if(target->user)
		addElement(cur, build_user(target->user));

	if(target->process)
		addElement(cur, build_process(target->process));

	if(target->service)
		addElement(cur, build_service(target->service));

	if(target->filelist)
		addElement(cur, build_filelist(target->filelist));

	return cur;
}

xmlNodePtr
build_classification(IDMEFclassification *classification) {
	xmlNodePtr cur = 0;

	cur = newClassification(NULL);

	if(classification->origin != -1)
		setAttribute(cur, newAttribute("origin",
		classification_origin_as_string(classification->origin)));

	if(classification->name)
		addElement(cur, newSimpleElement("name",
		classification->name));

	if(classification->url)
		addElement(cur, newSimpleElement("url",
		classification->url));
		
	return cur;
}

xmlNodePtr
build_analyzer(IDMEFanalyzer *analyzer) {
	xmlNodePtr cur = 0;

	cur = newAnalyzer(NULL);

	if(analyzer->analyzerid)
		setAttribute(cur, newAttribute("analyzerid",
			analyzer->analyzerid));

	if(analyzer->node)
		addElement(cur, build_node(analyzer->node));

	if(analyzer->process)
		addElement(cur, build_process(analyzer->process));

	return cur;
}

xmlNodePtr
build_impact(IDMEFimpact *impact) {
	xmlNodePtr cur = 0;

	cur = newImpact(newSimpleElement("value", impact->data), NULL);

	if(impact->severity != -1)
		setAttribute(cur, newAttribute("severity",
		impact_severity_as_string(impact->severity)));

	if(impact->completion != -1)
		setAttribute(cur, newAttribute("completion",
		impact_completion_as_string(impact->completion)));

	if(impact->type != -1)
		setAttribute(cur, newAttribute("type",
		impact_type_as_string(impact->type)));
	
	return cur;
}

xmlNodePtr
build_action(IDMEFaction *action) {
	xmlNodePtr cur = 0;

	cur = newAction(newSimpleElement("value", action->data), NULL);

	if(action->category != -1)
		setAttribute(cur, newAttribute("category",
		action_category_as_string(action->category)));

	return cur;
}

xmlNodePtr
build_confidence(IDMEFconfidence *confidence) {
	xmlNodePtr cur = 0;

	cur = newConfidence(newSimpleElement("value", confidence->data), NULL);

	if(confidence->rating != -1)
		setAttribute(cur, newAttribute("rating",
		confidence_rating_as_string(confidence->rating)));

	return cur;
}

xmlNodePtr
build_assessment(IDMEFassessment *assessment) {
	xmlNodePtr cur = 0;
	int i;

	cur = newAssessment(NULL);

	if(assessment->impact)
		addElement(cur, build_impact(assessment->impact));

	for(i = 0; i < assessment->nactions; i++) {
		addElement(cur, build_action(assessment->actions[i]));
	}

	if(assessment->confidence)
		addElement(cur, build_confidence(assessment->confidence));	

	return cur;
}

xmlNodePtr
build_toolalert(IDMEFtoolalert *toolalert) {
	xmlNodePtr cur = 0, child = 0;
	int i = 0;

	cur = newToolAlert(NULL);

	if(toolalert->name)
		addElement(cur, newSimpleElement("name",
		toolalert->name));

	if(toolalert->command)
		addElement(cur, newSimpleElement("command",
		toolalert->command));

	for(i = 0; i < toolalert->nalertidents; i++) {
		child = addElement(cur, newSimpleElement("alertident",
		toolalert->alertidents[i]->data));
		if(toolalert->alertidents[i]->analyzerid)
			setAttribute(child, newAttribute("analyzerid",
			toolalert->alertidents[i]->analyzerid));
	}

	return cur;
}

xmlNodePtr
build_overflowalert(IDMEFoverflowalert *overflowalert) {
	xmlNodePtr cur = 0;

	cur = newOverflowAlert(NULL);

	if(overflowalert->program)
		addElement(cur, newSimpleElement("program",
		overflowalert->program));

	if(overflowalert->size != -1) {
		char buf[BUFSIZ];
		memset(buf, '\0', sizeof(buf));
		sprintf(buf, "%d", overflowalert->size);
		addElement(cur, newSimpleElement("size", buf));
	}

	if(overflowalert->buffer)
		addElement(cur, newSimpleElement("buffer",
		overflowalert->buffer));

	return cur;
}

xmlNodePtr
build_correlationalert(IDMEFcorrelationalert *correlationalert) {
	xmlNodePtr cur = 0, child = 0;
	int i = 0;

	cur = newCorrelationAlert(NULL);

	if(correlationalert->name)
		addElement(cur, newSimpleElement("name",
		correlationalert->name));

	for(i = 0; i < correlationalert->nalertidents; i++) {
		child = addElement(cur, newSimpleElement("alertident",
		correlationalert->alertidents[i]->data));
		if(correlationalert->alertidents[i]->analyzerid)
			setAttribute(child, newAttribute("analyzerid",
			correlationalert->alertidents[i]->analyzerid));
	}

	return cur;
}

xmlNodePtr
build_additionaldata(IDMEFadditionaldata *additionaldata) {
	xmlNodePtr cur = 0;

	cur = newAdditionalData(
		newSimpleElement("value", additionaldata->data), NULL);

	if(additionaldata->meaning)
		setAttribute(cur, newAttribute("meaning",
		additionaldata->meaning));

	if(additionaldata->type != -1) 
		setAttribute(cur, newAttribute("type",
		additionaldata_type_as_string(additionaldata->type)));

	return cur;
}

xmlNodePtr
build_alert(IDMEFalert *alert) {
	xmlNodePtr cur = 0;
	int i = 0;

	cur = newAlert(NULL);

	if(alert->ident)
		setAttribute(cur, newAttribute("ident", alert->ident));

	if(alert->analyzer) {
		addElement(cur, build_analyzer(alert->analyzer));
	}

	if(alert->createtime) {
		addElement(cur, newCreateTime(&alert->createtime->tv));
	}

	if(alert->detecttime) {
		addElement(cur, newDetectTime(&alert->detecttime->tv));
	}

	if(alert->analyzertime) {
		addElement(cur, newAnalyzerTime(&alert->analyzertime->tv));
	}

	for(i = 0; i < alert->nsources; i++) {
		addElement(cur, build_source(alert->sources[i]));
	}

	for(i = 0; i < alert->ntargets; i++) {
		addElement(cur, build_target(alert->targets[i]));
	}

	for(i = 0; i < alert->nclassifications; i++) {
		addElement(cur,
			build_classification(alert->classifications[i]));
	}

	if(alert->assessment)
		addElement(cur, build_assessment(alert->assessment));

	if(alert->toolalert)
		addElement(cur,
			build_toolalert(alert->toolalert));

	if(alert->overflowalert)
		addElement(cur,
			build_overflowalert(alert->overflowalert));

	if(alert->correlationalert)
		addElement(cur,
			build_correlationalert(alert->correlationalert));

	for(i = 0; i < alert->nadditionaldatas; i++) {
		addElement(cur, build_additionaldata(
		alert->additionaldatas[i]));
	}

	return cur;
}

xmlNodePtr
build_heartbeat(IDMEFheartbeat *heartbeat) {
	xmlNodePtr cur = 0;
	int i = 0;

	cur = newHeartbeat(NULL);

	if(heartbeat->ident)
		setAttribute(cur, newAttribute("ident", heartbeat->ident));

	if(heartbeat->analyzer) {
		addElement(cur, build_analyzer(heartbeat->analyzer));
	}

	if(heartbeat->createtime) {
		addElement(cur, newCreateTime(&heartbeat->createtime->tv));
	}

	if(heartbeat->analyzertime) {
		addElement(cur, newAnalyzerTime(&heartbeat->analyzertime->tv));
	}

	for(i = 0; i < heartbeat->nadditionaldatas; i++) {
		addElement(cur, build_additionaldata(
		heartbeat->additionaldatas[i]));
	}

	return cur;
}

xmlNodePtr
build_idmef_message(IDMEFmessage *message) {
	xmlNodePtr cur = 0;
	int i;

	cur = newIDMEF_Message(NULL);

	if(message->version)
		setAttribute(cur, newAttribute("version", message->version));

	for(i = 0; i < message->nalerts; i++) {
		addElement(cur, build_alert(message->alerts[i]));
	}

	for(i = 0; i < message->nheartbeats; i++) {
		addElement(cur, build_heartbeat(message->heartbeats[i]));
	}

	return cur;
}

/*
** Name: generate_idmef_message
**
** Purpose: To generate an XML representation of the IDMEF message
** represented by the C structure passed as it's argument.
** 
** Arguments: The C structure containing the IDMEF message infomation.
**
** Returns: A character array containing the XML message.
**
** Notes:
**
*/

char *
generate_idmef_message(IDMEFmessage *message) {
	xmlDocPtr doc = 0;

	globalsInit(IDMEF_MESSAGE_DTD);

	createCurrentDoc(XML_VERSION);

	build_idmef_message(message);

	validateCurrentDoc();

	doc = getCurrentDoc();

	xmlDocDumpMemory(doc, (xmlChar **)&text, &text_length);

	return text;
}
