/*
idmefxml_parse.c: libidmef parsing functions
Author: Adam Migus NAI Labs (amigus@nai.com)

Copyright (c) 2001 Networks Associates Technology, Inc.
All rights reserved

This library is released under the GNU GPL and BSD software licenses.
You may choose to use one or the other, BUT NOT BOTH.  The GNU GPL
license is located in the file named COPYING.  The BSD license is located
in the file named COPYING.BSD.  Please contact us if there are any
questions.

*/

/*
This code assumes that message elements are in order as dictated by the
specification put forth by the IETF IDWG in the standards document
draft-ietf-idwg-idmef-xml-06.txt.  Attempting to parse messages with
elements out of order will result in undefined behavior.

$Id: idmefxml_parse.c,v 1.11 2002/02/21 22:37:58 amigus Exp $
*/

#include <stdio.h>
#include <string.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <stdlib.h>
#include <netdb.h>
#include "idmefxml_parse.h"

#ifdef IDMEF_DEBUG
#define DEBUG 1
#else
#define DEBUG 0
#endif /* IDMEF_DEBUG */

char * get_attribute(xmlNodePtr cur, char *name, int return_perm_string);
char * get_value(xmlNodePtr cur, int return_perm_string);
char ** get_values(xmlNodePtr cur, char *name, int *count);

/*
** Name: count_elements
**
** Purpose: Count the number of elements of a given name
**
** Arguments: The xmlNodePtr pointing to the elements you
** want to count.  The name of the element(s) you wish
** to count.
**
** Returns: The number of elements with the given name.
**
** Notes:
**
*/

int
count_elements(xmlNodePtr cur, char *name) {
	int n = 0;

	while(cur) {
		if(!strcasecmp(cur->name, name)) n++;
		cur = cur->next;
	}
	
	return n;
}

/*
** Name: get_attribute
**
** Purpose: Retrieves the named attribute from the attributes of
** the given xmlNodePtr.
**
** Arguments: The xmlNodePtr for which you want the attribute from.
** The string containing the name of the attribute you seek.
**
** Returns: A pointer to the string representing the attribute
** on success.  A null pointer on failure.
**
** Notes:
**
*/

char *
get_attribute(xmlNodePtr cur, char *name, int return_perm_string) {
	xmlAttrPtr attr;

	attr = cur->properties;

	while(attr && xmlStrcmp(name, attr->name)) attr = attr->next;

	if(!(attr)) return 0;

	if(!(attr->children)) return 0;

	if(!(attr->children->content)) return 0;

	if(return_perm_string) return strdup(attr->children->content);
	return attr->children->content;
}

/*
** Name: get_value
**
** Purpose: Retrieves the value of the given xmlNodePtr.
**
** Arguments: The xmlNodePtr for which you want the value from.
**
** Returns: A pointer to the string representing the value
** on success.  A null pointer on failure.
**
** Notes:
**
*/

char *
get_value(xmlNodePtr cur, int return_perm_string) {
	char *p;

	if(!(cur)) return 0;

	if(!(cur->children)) return 0;

	if(!(cur->children->content)) return 0;

	if(return_perm_string) {
		p = strdup(cur->children->content);
		return p;
	}	

	return cur->children->content;
}

/*
** Name: get_values
**
** Purpose: Retrieves the list of values of the given name starting
** at the xmlNodePtr given.
**
** Arguments: The xmlNodePtr pointing to the start of the xmlNodePtr's
** for which you want the values from.  The name of the values you
** want.  A pointer to the integer that will be set to the number
** of values returned.
**
** Returns: A pointer to an ALLOCATED array of string pointers, one
** for each value found on success.  A null pointer on failure.
**
** Notes:
**
** 1. It is up to the caller to free this allocated array when
** done with it.
**
*/

char **
get_values(xmlNodePtr cur, char *name, int *count) {
	char **values;
	int i = 0, n = 0;

	if(!(cur)) return 0;

	n = count_elements(cur, name);

	if(!n) return 0;

	values = (char **)malloc(n * sizeof(char *));
	memset(values, '\0', n * sizeof(char *));

	for(i = 0; i < n && cur; i++) {
		if(!strcasecmp(cur->name, name))
			values[i] = strdup(cur->children->content);
		cur = cur->next;
	}

	*count = n;

	return values;
}

/*
** The following section contains functions for translating between
** text and the inumerated types found in the header file.  I have
** written one comment block for all the functions.
*/

/*
** Name: Various
**
** Purpose: To convert between the inumerated types and their text
** equivalents.
**
** Arguments: Either a string for which the inumerated type is desired
** or an enumerated type for which the corresponding string is desired.
**
** Returns: Functions that take a string argument return the
** corresponding enumerated type or -1 on failure.  Functions that take
** an enumerated type return the corresponding string or null on failure.
**
** Notes:
**
*/

char *
node_category_as_string(IDMEFnode_category category) {
	switch(category) {
	case NODE_CATEGORY_UNKNOWN:
		return "unknown";
	case ADS:
		return "ads";
	case AFS:
		return "afs";
	case CODA:
		return "coda";
	case DFS:
		return "dfs";
	case DNS:
		return "dns";
	case KERBEROS:
		return "kerberos";
	case NDS:
		return "nds";
	case NIS:
		return "nis";
	case NISPLUS:
		return "nisplus";
	case NT:
		return "nt";
	case WFW:
		return "wfw";
	}

	return 0;
}

IDMEFnode_category
get_node_category(char *p) {
	if(!strcmp(p, "unknown")) {
		return NODE_CATEGORY_UNKNOWN;
	}
	else
	if(!strcmp(p, "ads")) {
		return ADS;
	}
	else
	if(!strcmp(p, "afs")) {
		return AFS;
	}
	else
	if(!strcmp(p, "coda")) {
		return CODA;
	}
	else
	if(!strcmp(p, "dfs")) {
		return DFS;
	}
	else
	if(!strcmp(p, "dns")) {
		return DNS;
	}
	else
	if(!strcmp(p, "kerberos")) {
		return KERBEROS;
	}
	else
	if(!strcmp(p, "nds")) {
		return NDS;
	}
	else
	if(!strcmp(p, "nis")) {
		return NIS;
	}
	else
	if(!strcmp(p, "nisplus")) {
		return NISPLUS;
	}
	else
	if(!strcmp(p, "nt")) {
		return NT;
	}
	else
	if(!strcmp(p, "wfw")) {
		return WFW;
	}

	return -1;
}

char *
address_category_as_string(IDMEFaddress_category category) {
	switch(category) {
	case ADDRESS_CATEGORY_UNKNOWN:
		return "unknown";
	case ATM:
		return "atm";
	case E_MAIL:
		return "e-mail";
	case LOTUS_NOTES:
		return "lotus-notes";
	case MAC:
		return "mac";
	case SNA:
		return "sna";
	case VM:
		return "vm";
	case IPV4_ADDR:
		return "ipv4-addr";
	case IPV4_ADDR_HEX:
		return "ipv4-addr-hex";
	case IPV4_NET:
		return "ipv4-net";
	case IPV4_NET_MASK:
		return "ipv4-net-mask";
	case IPV6_ADDR:
		return "ipv6-addr";
	case IPV6_ADDR_HEX:
		return "ipv6-addr-hex";
	case IPV6_NET:
		return "ipv6-net";
	case IPV6_NET_MASK:
		return "ipv6-net-mask";
	}

	return 0;
}

IDMEFaddress_category
get_address_category(char *p) {
	if(!strcmp(p, "unknown")) {
		return ADDRESS_CATEGORY_UNKNOWN;
	}
	else
	if(!strcmp(p, "atm")) {
		return ATM;
	}
	else
	if(!strcmp(p, "e-mail")) {
		return E_MAIL;
	}
	else
	if(!strcmp(p, "lotus-notes")) {
		return LOTUS_NOTES;
	}
	else
	if(!strcmp(p, "mac")) {
		return MAC;
	}
	else
	if(!strcmp(p, "sna")) {
		return SNA;
	}
	else
	if(!strcmp(p, "vm")) {
		return VM;
	}
	else
	if(!strcmp(p, "ipv4-addr")) {
		return IPV4_ADDR;
	}
	else
	if(!strcmp(p, "ipv4-addr-hex")) {
		return IPV4_ADDR_HEX;
	}
	else
	if(!strcmp(p, "ipv4-net")) {
		return IPV4_NET;
	}
	else
	if(!strcmp(p, "ipv4-net-mask")) {
		return IPV4_NET_MASK;
	}
	else
	if(!strcmp(p, "ipv6-addr")) {
		return IPV6_ADDR;
	}
	else
	if(!strcmp(p, "ipv6-addr-hex")) {
		return IPV6_ADDR_HEX;
	}
	else
	if(!strcmp(p, "ipv6-net")) {
		return IPV6_NET;
	}
	else
	if(!strcmp(p, "ipv6-net-mask")) {
		return IPV6_NET_MASK;
	}

	return -1;
}

IDMEFspoofed
get_spoofed(char *p) {
	if(!strcmp(p, "yes")) {
		return SPOOFED_YES;
	}
	else
	if(!strcmp(p, "no")) {
		return SPOOFED_NO;
	}
	else
	if(!strcmp(p, "unknown")) {
		return SPOOFED_UNKNOWN;
	}

	return -1;
}

char *
spoofed_as_string(IDMEFspoofed spoofed) {
	switch(spoofed) {
	case SPOOFED_YES:
		return "yes";
	case SPOOFED_NO:
		return "no";
	case SPOOFED_UNKNOWN:
		return "unknown";
	}

	return 0;
}

IDMEFdecoy
get_decoy(char *p) {
	if(!strcmp(p, "yes")) {
		return DECOY_YES;
	}
	else
	if(!strcmp(p, "no")) {
		return DECOY_NO;
	}
	else
	if(!strcmp(p, "unknown")) {
		return DECOY_UNKNOWN;
	}

	return -1;
}

char *
decoy_as_string(IDMEFdecoy decoy) {
	switch(decoy) {
	case DECOY_YES:
		return "yes";
	case DECOY_NO:
		return "no";
	case DECOY_UNKNOWN:
		return "unknown";
	}

	return 0;
}

IDMEFuser_category
get_user_category(char *p) {
	if(!strcmp(p, "unknown")) {
		return USER_CATEGORY_UNKNOWN;
	}
	else
	if(!strcmp(p, "application")) {
		return APPLICATION;
	}
	else
	if(!strcmp(p, "os-device")) {
		return OS_DEVICE;
	}

	return -1;
}

char *
user_category_as_string(IDMEFuser_category category) {
	switch(category) {
	case USER_CATEGORY_UNKNOWN:
		return "unknown";
	case APPLICATION:
		return "application";
	case OS_DEVICE:
		return "os-device";
	}

	return 0;
}

IDMEFUserId_type
get_userid_type(char *p) {
	if(!strcmp(p, "current-user")) {
		return CURRENT_USER;
	}
	else
	if(!strcmp(p, "original-user")) {
		return ORIGINAL_USER;
	}
	else
	if(!strcmp(p, "target-user")) {
		return TARGET_USER;
	}
	else
	if(!strcmp(p, "user-privs")) {
		return USER_PRIVS;
	}
	else
	if(!strcmp(p, "current-group")) {
		return CURRENT_GROUP;
	}
	else
	if(!strcmp(p, "group-privs")) {
		return GROUP_PRIVS;
	}
	else
	if(!strcmp(p, "other-privs")) {
		return OTHER_PRIVS;
	}

	return -1;
}

char *
userid_type_as_string(IDMEFUserId_type type) {
	switch(type) {
	case CURRENT_USER:
		return "current-user";
	case ORIGINAL_USER:
		return "original-user";
	case TARGET_USER:
		return "target-user";
	case USER_PRIVS:
		return "user-privs";
	case CURRENT_GROUP:
		return "current-group";
	case GROUP_PRIVS:
		return "group-privs";
	case OTHER_PRIVS:
		return "other-privs";
	}

	return 0;
}

IDMEForigin
get_classification_origin(char *p) {
	if(!strcmp(p, "unknown")) {
		return UNKNOWN_ORIGIN;
	}
	else
	if(!strcmp(p, "bugtraqid")) {
		return BUGTRAQID;
	}
	else
	if(!strcmp(p, "cve")) {
		return CVE;
	}
	else
	if(!strcmp(p, "vendor-specific")) {
		return VENDOR_SPECIFIC;
	}

	return -1;
}

char *classification_origin_as_string(IDMEForigin origin) {
	switch(origin) {
	case UNKNOWN_ORIGIN:
		return "unknown";
	case BUGTRAQID:
		return "bugtraqid";
	case CVE:
		return "cve";
	case VENDOR_SPECIFIC:
		return "vendor-specific";
	}

	return 0;
}

IDMEFadditionaldata_type
get_additionaldata_type(char *p) {
	if(!strcmp(p, "boolean")) {
		return BOOLEAN;
	}
	else
	if(!strcmp(p, "byte")) {
		return BYTE;
	}
	else
	if(!strcmp(p, "character")) {
		return CHARACTER;
	}
	else
	if(!strcmp(p, "date-time")) {
		return DATETIME;
	}
	else
	if(!strcmp(p, "integer")) {
		return INTEGER;
	}
	else
	if(!strcmp(p, "ntpstamp")) {
		return NTPSTAMP;
	}
	else
	if(!strcmp(p, "portlist")) {
		return PORTLIST;
	}
	else
	if(!strcmp(p, "real")) {
		return REAL;
	}
	else
	if(!strcmp(p, "string")) {
		return STRING;
	}
	else
	if(!strcmp(p, "xml")) {
		return XML;
	}

	return -1;
}

char *
additionaldata_type_as_string(IDMEFadditionaldata_type type) {
	switch(type) {
	case BOOLEAN:
		return "boolean";
	case BYTE:
		return "byte";
	case CHARACTER:
		return "character";
	case DATETIME:
		return "date-time";
	case INTEGER:
		return "integer";
	case NTPSTAMP:
		return "ntpstamp";
	case PORTLIST:
		return "portlist";
	case REAL:
		return "real";
	case STRING:
		return "string";
	case XML:
		return "xml";
	}

	return 0;
}

IDMEFimpact_severity
get_impact_severity(char *p) {
	if(!strcmp(p, "low")) {
		return IMPACT_SEVERITY_LOW;
	}
	else
	if(!strcmp(p, "medium")) {
		return IMPACT_SEVERITY_MEDIUM;
	}
	else
	if(!strcmp(p, "high")) {
		return IMPACT_SEVERITY_HIGH;
	}

	return -1;
}

char *
impact_severity_as_string(IDMEFimpact_severity severity) {
	switch(severity) {
	case IMPACT_SEVERITY_LOW:
		return "low";
	case IMPACT_SEVERITY_MEDIUM:
		return "medium";
	case IMPACT_SEVERITY_HIGH:
		return "high";
	}

	return 0;
}

IDMEFimpact_completion
get_impact_completion(char *p) {
	if(!strcmp(p, "failed")) {
		return FAILED;
	}
	else
	if(!strcmp(p, "succeeded")) {
		return SUCCEEDED;
	}

	return -1;
}

char *
impact_completion_as_string(IDMEFimpact_completion completion) {
	switch(completion) {
	case FAILED:
		return "failed";
	case SUCCEEDED:
		return "succeeded";
	}

	return 0;
}

IDMEFimpact_type
get_impact_type(char *p) {
	if(!strcmp(p, "admin")) {
		return ADMIN;
	}
	else
	if(!strcmp(p, "dos")) {
		return DOS;
	}
	else
	if(!strcmp(p, "file")) {
		return IMPACT_TYPE_FILE;
	}
	else
	if(!strcmp(p, "recon")) {
		return RECON;
	}
	else
	if(!strcmp(p, "user")) {
		return USER;
	}
	else
	if(!strcmp(p, "other")) {
		return IMPACT_TYPE_OTHER;
	}

	return -1;
}

char *
impact_type_as_string(IDMEFimpact_type type) {
	switch(type) {
	case ADMIN:
		return "admin";
	case DOS:
		return "dos";
	case IMPACT_TYPE_FILE:
		return "file";
	case RECON:
		return "recon";
	case USER:
		return "user";
	case IMPACT_TYPE_OTHER:
		return "other";
	}

	return 0;
}

IDMEFaction_category
get_action_category(char *p) {
	if(!strcmp(p, "other")) {
		return ACTION_CATEGORY_OTHER;
	}
	else
	if(!strcmp(p, "block-installed")) {
		return BLOCK_INSTALLED;
	}
	else
	if(!strcmp(p, "notification-sent")) {
		return NOTIFICATION_SENT;
	}
	else
	if(!strcmp(p, "taken-offline")) {
		return TAKEN_OFFLINE;
	}

	return -1;
}

char *
action_category_as_string(IDMEFaction_category category) {
	switch(category) {
	case ACTION_CATEGORY_OTHER:
		return "other";
	case BLOCK_INSTALLED:
		return "block-installed";
	case NOTIFICATION_SENT:
		return "notification-sent";
	case TAKEN_OFFLINE:
		return "taken-offline";
	}

	return 0;
}

IDMEFconfidence_rating
get_confidence_rating(char *p) {
	if(!strcmp(p, "low")) {
		return CONFIDENCE_RATING_LOW;
	}
	else
	if(!strcmp(p, "medium")) {
		return CONFIDENCE_RATING_MEDIUM;
	}
	else
	if(!strcmp(p, "high")) {
		return CONFIDENCE_RATING_HIGH;
	}
	else
	if(!strcmp(p, "numeric")) {
		return CONFIDENCE_RATING_NUMERIC;
	}

	return -1;
}

char *
confidence_rating_as_string(IDMEFconfidence_rating rating) {
	switch(rating) {
	case CONFIDENCE_RATING_LOW:
		return "low";
	case CONFIDENCE_RATING_MEDIUM:
		return "medium";
	case CONFIDENCE_RATING_HIGH:
		return "high";
	case CONFIDENCE_RATING_NUMERIC:
		return "numeric";
	}

	return 0;
}

IDMEFfile_category
get_file_category(char *p) {
	if(!strcmp(p, "current")) {
		return CURRENT;
	}
	else
	if(!strcmp(p, "original")) {
		return ORIGINAL;
	}

	return -1;
}

char *
file_category_as_string(IDMEFfile_category category) {
	switch(category) {
	case CURRENT:
		return "current";
	case ORIGINAL:
		return "original";
	}

	return 0;
}

IDMEFlinkage_category
get_linkage_category(char *p) {
	if(!strcmp(p, "hard-link")) {
		return HARD_LINK;
	}
	else
	if(!strcmp(p, "mount-point")) {
		return MOUNT_POINT;
	}
	else
	if(!strcmp(p, "reparse-point")) {
		return REPARSE_POINT;
	}
	else
	if(!strcmp(p, "shortcut")) {
		return SHORTCUT;
	}
	else
	if(!strcmp(p, "stream")) {
		return STREAM;
	}
	else
	if(!strcmp(p, "symbolic-link")) {
		return SYMBOLIC_LINK;
	}

	return -1;
}

char *
linkage_category_as_string(IDMEFlinkage_category category) {
	switch(category) {
	case HARD_LINK:
		return "hard-link";
	case MOUNT_POINT:
		return "mount-point";
	case REPARSE_POINT:
		return "reparse-point";
	case SHORTCUT:
		return "shortcut";
	case STREAM:
		return "stream";
	case SYMBOLIC_LINK:
		return "symbolic-link";
	}

	return 0;
}

/*
** The next section contains the parsing routines.  All of these
** functions take an xmlNodePtr pointing to the element it is
** to parse and a pointer to the C structure representing that
** element type.
** These parsing routines can be used by themselves by the user
** if need be but for normal operation the user need only call
** the get_idmef_message() or get_idmef_message_from_file()
** functions below which internally call each parsing routine
** upon encountering an element of that type in the message.
** I include one comment block for all the functions
*/

/*
** Name: parse_XXX where XXX is an element type defined in the
** IDMEF message exchange format specification.  Types are in
** lowercase in all cases.
**
** Purpose: The parse_XXX function parses the element XXX and
** populates it's corresponding C structure as defined in the
** header file.
**
** Arguments: In all cases the first argument is an xmlNodePtr
** that is assumed to contain the element XXX.  The second
** argument is always the corresponding C structure for the
** element type XXX.
**
** Returns: All functions return 1 on success, 0 on failure.
**
** Notes:
**
** 1. Functions that encounter ** arrays in the C structures
** they populate will ALLOCATE space as nessesary.  The user
** is responsible for freeing space allocated by these
** functions when done.  For convienience there is a series
** of free_XXX() functions which will free all occurences of
** such arrays that appear in a the associated type.
**
** 2. The functions are somewhat recursive in nature.  While
** none of them are truely recursive (no function ever calls
** itself) one of the parsing functions will call another
** when a child node of that type is found.  It will pass a
** pointer to the C structure for that corresponding element
** type.  Obviously a function will only call another function
** of a type that appears as a child of the calling function
** type.
**
** 3. All functions expect the child elements to be present
** when they are required to be there by the specification!
**
** 4. All functions expect that child elements to be in the
** order that is dictated by the specification.
**
** 5. Attempting to parse messages that fail to comply with
** notes 3 and 4 will result in undefined behavior!
*/

int
parse_userid(xmlNodePtr cur, IDMEFuserid *userid) {
	char *p = 0;

	userid->ident = get_attribute(cur, "ident",1);

	p = get_attribute(cur, "type",0);

	if(p) userid->type = get_userid_type(p);
	else userid->type = ORIGINAL_USER;

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "name")) {
		userid->name = get_value(cur,1);
		cur = cur->next;
	}

	if(!cur) return 1;

	if(!strcasecmp(cur->name, "number")) {
		p = get_value(cur,0);
		if(p) userid->number = atoi(p);
		cur = cur->next;
	}

	return 1;
}

int
parse_user(xmlNodePtr cur, IDMEFuser *user) {
	char *p = 0;
	int i, n = 0;
	
	user->ident = get_attribute(cur, "ident",1);
	
	p = get_attribute(cur, "category",0);

	if(p) user->category = get_user_category(p);
	else user->category = USER_CATEGORY_UNKNOWN;

	cur = cur->children;
	if(!cur) return 0;

	n = count_elements(cur, "UserId");

	if(n) {
		user->userids =
		(IDMEFuserid **)malloc(n * sizeof(IDMEFuserid *));
		memset(user->userids, '\0', n * sizeof(IDMEFuserid *));

		user->nuserids = n;


		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_user) next element is %s -- calling parse_userid()\n",cur->name);
			user->userids[i] =
			(IDMEFuserid *)malloc(sizeof(IDMEFuserid));
			memset(user->userids[i], '\0', sizeof(IDMEFuserid));
			if(!parse_userid(cur, user->userids[i])) {
				fprintf(stderr, "error parsing userid\n");
				return 0;
			}
			cur = cur->next;
		}
	}
	else return 0;

	return 1;
}

int
parse_time(xmlNodePtr cur, IDMEFtime *time, char *name) {
	char *start = 0, *end = 0;
	double d = 0.0;
	

	time->ntpstamp = get_attribute(cur, "ntpstamp",1);

	if(!time->ntpstamp) return 0;

	time->string = get_value(cur,1);

	start = time->ntpstamp;

	time->tv.tv_sec = strtoul(start, &end, 16) - 2208988800UL;

	start = ++end;

	d = ((double)strtoul(start, 0, 16) * 1000000.0) / 4294967296;

	time->tv.tv_usec = d;

	if((d - (double)time->tv.tv_usec) > 0.5) time->tv.tv_usec++;

	return 1;
}

int
parse_address(xmlNodePtr cur, IDMEFaddress *address) {
	char *p = 0;

	address->ident = get_attribute(cur, "ident",1);

	address->vlan_name = get_attribute(cur, "vlan-name",1);

	p = get_attribute(cur, "vlan-num",0);

	if(p) address->vlan_num = atoi(p);

	p = get_attribute(cur, "category",0);

	if(p) address->category = get_address_category(p);
	else address->category = ADDRESS_CATEGORY_UNKNOWN;

	cur = cur->children;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "address")) return 0;

	address->address = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 1;

	address->netmask = get_value(cur,1);

	return 1;
}

int
parse_process(xmlNodePtr cur, IDMEFprocess *process) {
	char *p = 0;
	int i = 0;

	process->ident = get_attribute(cur, "ident",1);

	cur = cur->children;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "name")) return 0;

	process->name = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 1;

	p = get_value(cur,0);

	if(!strcasecmp(cur->name, "pid")) {
		p = get_value(cur,0);
		if(p) process->pid = atoi(p);
		cur = cur->next;
	}

	cur = cur->next;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "path"))
		process->path = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 1;

	process->arg = get_values(cur, "arg", &process->narg);

	for(i = 0; i < process->narg; i++) cur = cur->next;

	process->env = get_values(cur, "env", &process->nenv);

	for(i = 0; i < process->nenv; i++) cur = cur->next;

	return 1;
}

int
parse_additionaldata(xmlNodePtr cur, IDMEFadditionaldata *additionaldata) {
	char *p = 0;

	p = get_attribute(cur, "type",0);

	if(p) additionaldata->type = get_additionaldata_type(p);
	else additionaldata->type = STRING;

	additionaldata->meaning = get_attribute(cur, "meaning",1);

	additionaldata->data = get_value(cur,1);

	return 1;
}

int
parse_classification(xmlNodePtr cur, IDMEFclassification *classification) {
	char *p = 0;

	p = get_attribute(cur, "origin",0);

	if(p) classification->origin = get_classification_origin(p);
	else classification->origin = -1;

	cur = cur->children;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "name")) return 0;

	classification->name = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "url")) return 0;

	classification->url = get_value(cur,1);

	return 1;
}

int
parse_impact(xmlNodePtr cur, IDMEFimpact *impact) {
	char *p = 0;

	p = get_attribute(cur, "severity",0);

	if(p) impact->severity = get_impact_severity(p);
	else impact->severity = -1;
	
	p = get_attribute(cur, "completion",0);

	if(p) impact->completion = get_impact_completion(p);
	else impact->completion = -1;

	p = get_attribute(cur, "type",0);

	if(p) impact->type = get_impact_type(p);
	else impact->type = IMPACT_TYPE_OTHER;

	impact->data = get_value(cur,1);

	return 1;
}

int
parse_action(xmlNodePtr cur, IDMEFaction *action) {
	char *p = 0;

	p = get_attribute(cur, "category",0);

	if(p) action->category = get_action_category(p);
	else action->category = ACTION_CATEGORY_OTHER;

	action->data = get_value(cur,1);

	return 1;
};

int
parse_confidence(xmlNodePtr cur, IDMEFconfidence *confidence) {
	char *p = 0;

	p = get_attribute(cur, "rating",0);

	if(p) confidence->rating = get_confidence_rating(p);
	else confidence->rating = CONFIDENCE_RATING_NUMERIC;

	confidence->data = get_value(cur,1);

	return 1;
};

int
parse_assessment(xmlNodePtr cur, IDMEFassessment *assessment) {
	int n = 0, i = 0;

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Impact")) {
		if(DEBUG) fprintf(stdout, "(parse_assessment) next element is %s -- calling parse_impact()\n",cur->name);
		assessment->impact = (IDMEFimpact *)malloc(sizeof(IDMEFimpact));
		memset(assessment->impact, '\0', sizeof(IDMEFimpact));
		if(!parse_impact(cur, assessment->impact)) {
			fprintf(stderr, "error parsing Impact\n");
			return 0;
		}
		cur = cur->next;
	}

	if(!cur) return 1;

	n = count_elements(cur, "Action");

	if(n) {
		assessment->actions =
		(IDMEFaction **)malloc(n * sizeof(IDMEFaction *));
		memset(assessment->actions, '\0',
		n * sizeof(IDMEFaction *));

		assessment->nactions = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_assessment) next element is %s -- calling parse_action()\n",cur->name);
			assessment->actions[i] =
			(IDMEFaction *)malloc(sizeof(IDMEFaction));
			memset(assessment->actions[i], '\0',
			sizeof(IDMEFaction));
			if(!parse_action(cur, assessment->actions[i])) {
				fprintf(stderr,
				"error parsing Action\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Confidence")) {
		if(DEBUG) fprintf(stdout, "(parse_assessment) next element is %s -- calling parse_confidence()\n",cur->name);
		assessment->confidence = (IDMEFconfidence *)malloc(sizeof(IDMEFconfidence));
		memset(assessment->confidence, '\0', sizeof(IDMEFconfidence));
		if(!parse_confidence(cur, assessment->confidence)) {
			fprintf(stderr, "error parsing Confidence\n");
			return 0;
		}
		cur = cur->next;
	}

	return 1;
}

int
parse_node(xmlNodePtr cur, IDMEFnode *node) {
	char *p = 0;
	int i = 0, n = 0;

	node->ident = get_attribute(cur, "ident",1);

	p = get_attribute(cur, "category",0);

	if(p) node->category = get_node_category(p);
	else node->category = NODE_CATEGORY_UNKNOWN;

	cur = cur->children;
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "location")) {
		node->location = get_value(cur,1);
		cur = cur->next;
	}

	if(!strcasecmp(cur->name, "name")) {
		node->name = get_value(cur,1);
		cur = cur->next;
	}

	if(!cur) return 1;

	n = count_elements(cur, "Address");

	if(n) {
		node->addresses =
		(IDMEFaddress **)malloc(n * sizeof(IDMEFaddress *));
		memset(node->addresses, '\0',
		n * sizeof(IDMEFaddress *));

		node->naddresses = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_node) next element is %s -- calling parse_address()\n",cur->name);
			node->addresses[i] =
			(IDMEFaddress *)malloc(sizeof(IDMEFaddress));
			memset(node->addresses[i], '\0',
			sizeof(IDMEFaddress));
			if(!parse_address
			(cur, node->addresses[i])) {
				fprintf(stderr,
				"error parsing Address\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	if(node->naddresses < 1 && !node->name) return 0;

	return 1;
}

int
parse_webservice(xmlNodePtr cur, IDMEFwebservice *webservice) {
	cur = cur->children;
	if(!cur) return 0;
	
	if(!strcasecmp(cur->name, "url")) {
		webservice->url = get_value(cur,1);
	}
	else return 0;

	cur = cur->next;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "cgi")) {
		webservice->cgi = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "http-method")) {
		webservice->http_method = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}

	webservice->arg = get_values(cur, "arg", &webservice->narg);

	return 1;
}

int
parse_snmpservice(xmlNodePtr cur, IDMEFSNMPservice *snmpservice) {
	cur = cur->children;

	if(!cur) return 1;

	if(!strcasecmp(cur->name, "oid")) {
		snmpservice->oid = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "community")) {
		snmpservice->community = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "command")) {
		snmpservice->command = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}

	return 1;
}

int
parse_service(xmlNodePtr cur, IDMEFservice *service) {
	struct protoent *pr;
	char * p = 0;

	service->ident = get_attribute(cur, "ident",1);

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "name")) {
		service->name = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;

	}

	if(!strcasecmp(cur->name, "port")) {
		p = get_value(cur,0);
		if(p) service->port = atoi(p);
		cur = cur->next;
		if(!cur) return 1;

	}

	if(!strcasecmp(cur->name, "portlist")) {
		service->portlist = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!service->name && !service->port && !service->portlist) return 0;

	if(!strcasecmp(cur->name, "protocol")) {
		p = get_value(cur,0);
		pr = getprotobyname(p);
		if(!pr) {
			fprintf(stderr, "invalid protocol: %s\n", p);
			return 0;
		}
		service->protocol = pr->p_proto;

		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "SNMPService")) {
		if(DEBUG) fprintf(stdout, "(parse_service) next element is %s -- calling parse_snmpservice()\n",cur->name);
		service->snmpservice = (IDMEFSNMPservice *)malloc(sizeof(IDMEFSNMPservice));
		memset(service->snmpservice, '\0', sizeof(IDMEFSNMPservice));
		if(!parse_snmpservice(cur, service->snmpservice)) {
			fprintf(stderr, "error parsing SNMPService\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "WebService")) {
		if(DEBUG) fprintf(stdout, "(parse_service) next element is %s -- calling parse_webservice()\n",cur->name);
		service->webservice = (IDMEFwebservice *)malloc(sizeof(IDMEFwebservice));
		memset(service->webservice, '\0', sizeof(IDMEFwebservice));
		if(!parse_webservice(cur, service->webservice)) {
			fprintf(stderr, "error parsing WebService\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	return 1;
}

int
parse_linkage(xmlNodePtr cur, IDMEFlinkage *linkage) {
	char *p;

	p = get_attribute(cur, "category", 0);
	if(p) linkage->category = get_linkage_category(p);

	cur = cur->children;
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "name")) {
		linkage->name = get_value(cur, 1);
		cur = cur->next;
	}

	if(!cur) return 0;

	if(!strcasecmp(cur->name, "path")) {
		linkage->path = get_value(cur, 1);
		cur = cur->next;
	}

	if(linkage->name && linkage->path) {
		if(!cur) return 1;
		else return 0;
	}

	if(!cur) return 0;

	if(!strcasecmp(cur->name, "File")) {
		if(!parse_file(cur, linkage->file)) {
			fprintf(stderr, "error parsing File\n");
			return 0;
		}
		cur = cur->next;
	}

	if(linkage->file) return 1;

	if(!linkage->name || !linkage->path) return 0;

	return 1;
}

int
parse_inode(xmlNodePtr cur, IDMEFinode *inode) {
	char *p = 0;
	int got_major = 0, got_minor = 0;

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "change-time")) {
		inode->change_time = get_value(cur, 1);
		cur = cur->next;
	}

	if(!cur) return 1;

	if(!strcasecmp(cur->name, "number")) {
		p = get_value(cur, 0);
		inode->number = atoi(p);
		cur = cur->next;
	}
	
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "major-device")) {
		p = get_value(cur, 0);
		inode->major_device = atoi(p);
		got_major = 1;
		cur = cur->next;
	}

	if(!cur) return 0;

	if(!strcasecmp(cur->name, "minor-device")) {
		p = get_value(cur, 0);
		inode->minor_device = atoi(p);
		got_minor = 1;
		cur = cur->next;
	}

	if(inode->number && !(got_major || got_minor)) return 0;

	if(!cur) return 1;

	got_major = got_minor = 0;

	if(!strcasecmp(cur->name, "c-major-device")) {
		p = get_value(cur, 0);
		inode->c_major_device = atoi(p);
		got_major = 1;
		cur = cur->next;
	}

	if(!cur) return 0;

	if(!strcasecmp(cur->name, "c-minor-device")) {
		p = get_value(cur, 0);
		inode->c_minor_device = atoi(p);
		got_minor = 1;
		cur = cur->next;
	}
	
	if(!(got_major || got_minor)) return 0;

	return 1;
}

int
parse_fileaccess(xmlNodePtr cur, IDMEFfileaccess *fileaccess) {
	int i;

	cur = cur->children;
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "UserId")) {
		if(DEBUG) fprintf(stdout, "(parse_fileaccess) next element is %s -- calling parse_userid()\n",cur->name);
		fileaccess->userid = (IDMEFuserid *)malloc(sizeof(IDMEFuserid));
		memset(fileaccess->userid, '\0', sizeof(IDMEFuserid));
		if(!parse_userid(cur, fileaccess->userid)) {
			fprintf(stderr, "error parsing UserId\n");
			return 0;
		}
		cur = cur->next;
	}

	if(!cur) return 0;

	if(strcmp(
	userid_type_as_string(fileaccess->userid->type), "user-privs") &&
	strcmp(userid_type_as_string(fileaccess->userid->type), "group-privs") &&
	strcmp(userid_type_as_string(fileaccess->userid->type), "other-privs"))
		return 0;

	fileaccess->permissions = get_values(cur, "permission",
	&fileaccess->npermissions);

	for(i = 0; i < fileaccess->npermissions; i++) cur = cur->next;

	if(!fileaccess->npermissions) return 0;

	return 1;
}

int
parse_file(xmlNodePtr cur, IDMEFfile *file) {
	int i = 0, n = 0;
	char *p = 0;
	
	file->ident = get_attribute(cur, "ident",1);

	p = get_attribute(cur, "category", 0);
	if(!p) return 0;
	file->category = get_file_category(p);

	file->fstype = get_attribute(cur, "fstype",1);
	if(!file->fstype) return 0;

	cur = cur->children;
	if(!cur) return 0;

	file->name = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 0;

	file->path = get_value(cur,1);
	cur = cur->next;
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "create-time")) {
		file->create_time = get_value(cur, 1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "modify-time")) {
		file->modify_time = get_value(cur, 1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "access-time")) {
		file->access_time = get_value(cur, 1);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "data-size")) {
		p = get_value(cur,0);
		file->data_size = atoi(p);
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "disk-size")) {
		p = get_value(cur,0);
		file->disk_size = atoi(p);
		cur = cur->next;
		if(!cur) return 1;
	}

	n = count_elements(cur, "FileAccess");

	if(n) {
		file->fileaccesses =
		(IDMEFfileaccess **)malloc(n * sizeof(IDMEFfileaccess *));
		memset(file->fileaccesses, '\0',
		n * sizeof(IDMEFfileaccess *));

		file->nfileaccesses = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_file) next element is %s -- calling parse_fileaccess()\n",cur->name);
			file->fileaccesses[i] =
			(IDMEFfileaccess *)malloc(sizeof(IDMEFfileaccess));
			memset(file->fileaccesses[i], '\0',
			sizeof(IDMEFfileaccess));
			if(!parse_fileaccess (cur, file->fileaccesses[i])) {
				fprintf(stderr, "error parsing FileAccess\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	if(!cur) return 1;

	n = count_elements(cur, "Linkage");

	if(n) {
		file->linkages =
		(IDMEFlinkage **)malloc(n * sizeof(IDMEFlinkage *));
		memset(file->linkages, '\0',
		n * sizeof(IDMEFlinkage *));

		file->nlinkages = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_file) next element is %s -- calling parse_linkage()\n",cur->name);
			file->linkages[i] =
			(IDMEFlinkage *)malloc(sizeof(IDMEFlinkage));
			memset(file->linkages[i], '\0',
			sizeof(IDMEFlinkage));
			if(!parse_linkage (cur, file->linkages[i])) {
				fprintf(stderr, "error parsing FileAccess\n");
				return 0;
			}
			cur = cur->next;
		}
	}
	
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Inode")) {
		if(DEBUG) fprintf(stdout, "(parse_file) next element is %s -- calling parse_inode()\n",cur->name);
		file->inode = (IDMEFinode *)malloc(sizeof(IDMEFinode));
		memset(file->inode, '\0', sizeof(IDMEFinode));
		if(!parse_inode(cur, file->inode)) {
			fprintf(stderr, "error parsing Inode\n");
			return 0;
		}
		cur = cur->next;
	}

	return 1;
}

int
parse_filelist(xmlNodePtr cur, IDMEFfilelist *filelist) {
	int i = 0, n = 0;

	cur = cur->children;
	if(!cur) return 0;

	n = count_elements(cur, "File");

	if(n) {
		filelist->files =
		(IDMEFfile **)malloc(n * sizeof(IDMEFfile *));
		memset(filelist->files, '\0',
		n * sizeof(IDMEFfile *));

		filelist->nfiles = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_filelist) next element is %s -- calling parse_file()\n",cur->name);
			filelist->files[i] =
			(IDMEFfile *)malloc(sizeof(IDMEFfile));
			memset(filelist->files[i], '\0', sizeof(IDMEFfile));
			if(!parse_file (cur, filelist->files[i])) {
				fprintf(stderr, "error parsing FileList\n");
				return 0;
			}
			cur = cur->next;
		}
	}
	else return 0;

	return 1;
}

int
parse_target(xmlNodePtr cur, IDMEFtarget *target) {
	char *p = 0;
	
	target->ident = get_attribute(cur, "ident",1);

	p = get_attribute(cur, "decoy",0);

	if(p) target->decoy = get_decoy(p);
	else target->decoy = -1;
		
	target->interface = get_attribute(cur, "interface",1);

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Node")) {
		if(DEBUG) fprintf(stdout, "(parse_target) next element is %s -- calling parse_node()\n",cur->name);
		target->node = (IDMEFnode *)malloc(sizeof(IDMEFnode));
		memset(target->node, '\0', sizeof(IDMEFnode));
		if(!parse_node(cur, target->node)) {
			fprintf(stderr, "error parsing Node\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "User")) {
		if(DEBUG) fprintf(stdout, "(parse_target) next element is %s -- calling parse_user()\n",cur->name);
		target->user = (IDMEFuser *)malloc(sizeof(IDMEFuser));
		memset(target->user, '\0', sizeof(IDMEFuser));
		if(!parse_user(cur, target->user)) {
			fprintf(stderr, "error parsing User\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "Process")) {
		if(DEBUG) fprintf(stdout, "(parse_target) next element is %s -- calling parse_process()\n",cur->name);
		target->process = (IDMEFprocess *)malloc(sizeof(IDMEFprocess));
		memset(target->process, '\0', sizeof(IDMEFprocess));
		if(!parse_process(cur, target->process)) {
			fprintf(stderr, "error parsing Process\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "Service")) {
		if(DEBUG) fprintf(stdout, "(parse_target) next element is %s -- calling parse_service()\n",cur->name);
		target->service = (IDMEFservice *)malloc(sizeof(IDMEFservice));
		memset(target->service, '\0', sizeof(IDMEFservice));
		if(!parse_service(cur, target->service)) {
			fprintf(stderr, "error parsing Service\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "FileList")) {
		if(DEBUG) fprintf(stdout, "(parse_target) next element is %s -- calling parse_filelist()\n",cur->name);
		target->filelist = (IDMEFfilelist *)
		malloc(sizeof(IDMEFfilelist));
		memset(target->filelist, '\0', sizeof(IDMEFfilelist));
		if(!parse_filelist(cur, target->filelist)) {
			fprintf(stderr, "error parsing FileList\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	return 1;
}

int
parse_source(xmlNodePtr cur, IDMEFsource *source) {
	char *p = 0;
	
	source->ident = get_attribute(cur, "ident",1);

	p = get_attribute(cur, "spoofed",0);

	if(p) source->spoofed = get_spoofed(p);
	else source->spoofed = -1;
		
	source->interface = get_attribute(cur, "interface",1);

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Node")) {
		if(DEBUG) fprintf(stdout, "(parse_source) next element is %s -- calling parse_node()\n",cur->name);
		source->node = (IDMEFnode *)malloc(sizeof(IDMEFnode));
		memset(source->node, '\0', sizeof(IDMEFnode));
		if(!parse_node(cur, source->node)) {
			fprintf(stderr, "error parsing Node\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "User")) {
		if(DEBUG) fprintf(stdout, "(parse_source) next element is %s -- calling parse_user()\n",cur->name);
		source->user = (IDMEFuser *)malloc(sizeof(IDMEFuser));
		memset(source->user, '\0', sizeof(IDMEFuser));
		if(!parse_user(cur, source->user)) {
			fprintf(stderr, "error parsing User\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "Process")) {
		if(DEBUG) fprintf(stdout, "(parse_source) next element is %s -- calling parse_process()\n",cur->name);
		source->process = (IDMEFprocess *)malloc(sizeof(IDMEFprocess));
		memset(source->process, '\0', sizeof(IDMEFprocess));
		if(!parse_process(cur, source->process)) {
			fprintf(stderr, "error parsing Process\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "Service")) {
		if(DEBUG) fprintf(stdout, "(parse_source) next element is %s -- calling parse_service()\n",cur->name);
		source->service = (IDMEFservice *)malloc(sizeof(IDMEFservice));
		memset(source->service, '\0', sizeof(IDMEFservice));
		if(!parse_service(cur, source->service)) {
			fprintf(stderr, "error parsing Service\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	return 1;
}

int
parse_analyzer(xmlNodePtr cur, IDMEFanalyzer *analyzer) {
	char *p = 0;

	p = get_attribute(cur, "analyzerid",1);

	if(p) analyzer->analyzerid = p;

	p = get_attribute(cur, "manufacturer",1);

	if(p) analyzer->manufacturer = p;

	p = get_attribute(cur, "model",1);

	if(p) analyzer->model = p;

	p = get_attribute(cur, "version",1);

	if(p) analyzer->version = p;

	p = get_attribute(cur, "class",1);

	if(p) analyzer->class = p;

	p = get_attribute(cur, "ostype",1);

	if(p) analyzer->ostype = p;

	p = get_attribute(cur, "osversion",1);

	if(p) analyzer->osversion = p;

	cur = cur->children;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Node")) {
		if(DEBUG) fprintf(stdout, "(parse_analyzer) next element is %s -- calling parse_node()\n",cur->name);
		analyzer->node = (IDMEFnode *)malloc(sizeof(IDMEFnode));
		memset(analyzer->node, '\0', sizeof(IDMEFnode));
		if(!parse_node(cur, analyzer->node)) {
			fprintf(stderr, "error parsing Node\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "Process")) {
		if(DEBUG) fprintf(stdout, "(parse_analyzer) next element is %s -- calling parse_process()\n",cur->name);
		analyzer->process = (IDMEFprocess *)malloc(sizeof(IDMEFprocess));
		memset(analyzer->process, '\0', sizeof(IDMEFprocess));
		if(!parse_process(cur, analyzer->process)) {
			fprintf(stderr, "error parsing Process\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}

	return 1;
}

int
parse_overflowalert(xmlNodePtr cur, IDMEFoverflowalert *overflowalert) {
	char *p = 0;

	cur = cur->children;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "program")) return 0;

	overflowalert->program = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "size")) {
		p = get_value(cur,0);
		if(p) overflowalert->size = atoi(p);
		else overflowalert->size = -1;
		cur = cur->next;
		if(!cur) return 1;
	}

	if(!strcasecmp(cur->name, "buffer")) {
		overflowalert->buffer = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 1;
	}


	return 1;
}

int
parse_toolalert(xmlNodePtr cur, IDMEFtoolalert *toolalert) {
	IDMEFalertident **idents;
	int i = 0, n = 0;

	cur = cur->children;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "name")) return 0;

	toolalert->name = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "command")) {
		toolalert->command = get_value(cur,1);
		cur = cur->next;
		if(!cur) return 0;
	}

	n = count_elements(cur, "alertident");

	if(!n) return 0;

	idents = (IDMEFalertident **)malloc(n * sizeof(IDMEFalertident *));
	memset(idents, '\0', n * sizeof(IDMEFalertident *));

	for(i = 0; i < n && cur; i++) {
		if(!strcasecmp(cur->name, "alertident")) {
			idents[i] =
			(IDMEFalertident *)malloc(sizeof(IDMEFalertident));
			idents[i]->data = get_value(cur, 1);
			idents[i]->analyzerid =
			get_attribute(cur, "analyzerid",1);
		}

		cur = cur->next;
	}

	toolalert->nalertidents = n;
	toolalert->alertidents = idents;

	return 1;
}

int
parse_correlationalert(xmlNodePtr cur,
			IDMEFcorrelationalert *correlationalert) {
	IDMEFalertident **idents;
	int i = 0, n = 0;

	cur = cur->children;
	if(!cur) return 0;

	if(strcasecmp(cur->name, "name")) return 0;

	correlationalert->name = get_value(cur,1);

	cur = cur->next;
	if(!cur) return 0;

	n = count_elements(cur, "alertident");

	if(!n) return 0;

	idents = (IDMEFalertident **)malloc(n * sizeof(IDMEFalertident *));
	memset(idents, '\0', n * sizeof(IDMEFalertident *));

	for(i = 0; i < n && cur; i++) {
		if(!strcasecmp(cur->name, "alertident")) {
			idents[i] =
			(IDMEFalertident *)malloc(sizeof(IDMEFalertident));
			idents[i]->data = get_value(cur, 1);
			idents[i]->analyzerid =
			get_attribute(cur, "analyzerid",1);
		}

		cur = cur->next;
	}

	correlationalert->nalertidents = n;
	correlationalert->alertidents = idents;

	return 1;
}

int
parse_heartbeat(xmlNodePtr cur, IDMEFheartbeat *heartbeat) {
	int n = 0, i = 0;

	heartbeat->ident = get_attribute(cur, "ident",1);

	cur = cur->children;
	if(!cur) return 0;

	/* Analyzer must appear here */
	if(strcasecmp(cur->name, "Analyzer")) return 0;

	if(DEBUG) fprintf(stdout, "(parse_heartbeat) next element is %s -- calling parse_analyzer()\n",cur->name);
	heartbeat->analyzer = (IDMEFanalyzer *)malloc(sizeof(IDMEFanalyzer));
	memset(heartbeat->analyzer, '\0', sizeof(IDMEFanalyzer));
	if(!parse_analyzer(cur, heartbeat->analyzer)) {
		fprintf(stderr, "error parsing Analyzer\n");
		return 0;
	}

	/* must return 0 since we didn't see CreateTime */
	cur = cur->next;
	if(!cur) return 0;

	/* CreateTime must appear here */
	if(strcasecmp(cur->name, "CreateTime")) return 0;

	if(DEBUG) fprintf(stdout, "(parse_heartbeat) next element is %s -- calling parse_time()\n",cur->name);
	heartbeat->createtime = (IDMEFtime *)malloc(sizeof(IDMEFtime));
	memset(heartbeat->createtime, '\0', sizeof(IDMEFtime));
	parse_time(cur, heartbeat->createtime, "CreateTime");

	cur = cur->next;
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "AnalyzerTime")) {
		if(DEBUG) fprintf(stdout, "(parse_heartbeat) next element is %s -- calling parse_time()\n",cur->name);
		heartbeat->analyzertime = (IDMEFtime *)malloc(sizeof(IDMEFtime));
		memset(heartbeat->analyzertime, '\0', sizeof(IDMEFtime));
		if(!parse_time(cur, heartbeat->analyzertime, "AnalyzerTime")) {
			fprintf(stderr, "error parsing AnalyzerTime\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 1;
	}
		
	/* If we get here it must be AdditionalData */
	if(strcasecmp(cur->name, "AdditionalData")) return 0;

	n = count_elements(cur, "AdditionalData");

	if(n) {
		heartbeat->additionaldatas =
		(IDMEFadditionaldata **)malloc(n * sizeof(IDMEFadditionaldata *));
		memset(heartbeat->additionaldatas, '\0',
		n * sizeof(IDMEFadditionaldata *));

		heartbeat->nadditionaldatas = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_heartbeat) next element is %s -- calling parse_additionaldata()\n",cur->name);
			heartbeat->additionaldatas[i] =
			(IDMEFadditionaldata *)malloc(sizeof(IDMEFadditionaldata));
			memset(heartbeat->additionaldatas[i], '\0',
			sizeof(IDMEFadditionaldata));
			if(!parse_additionaldata
			(cur, heartbeat->additionaldatas[i])) {
				fprintf(stderr,
				"error parsing AdditionalData\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	return 1;
}

int
parse_alert(xmlNodePtr cur, IDMEFalert *alert) {
	int n = 0, i = 0;

	alert->ident = get_attribute(cur, "ident",1);

	cur = cur->children;
	if(!cur) return 0;

	/* Analyzer must appear here */
	if(strcasecmp(cur->name, "Analyzer")) return 0;

	if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_analyzer()\n",cur->name);
	alert->analyzer = (IDMEFanalyzer *)malloc(sizeof(IDMEFanalyzer));
	memset(alert->analyzer, '\0', sizeof(IDMEFanalyzer));
	if(!parse_analyzer(cur, alert->analyzer)) {
		fprintf(stderr, "error parsing Analyzer\n");
		return 0;
	}

	cur = cur->next;
	/* must return 0 since we didn't see CreateTime */
	if(!cur) return 0;

	/* CreateTime must appear here */
	if(strcasecmp(cur->name, "CreateTime")) return 0;

	if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_time()\n",cur->name);
	alert->createtime = (IDMEFtime *)malloc(sizeof(IDMEFtime));
	memset(alert->createtime, '\0', sizeof(IDMEFtime));
	parse_time(cur, alert->createtime, "CreateTime");

	cur = cur->next;
	/* must return 0 since we didn't see Classification */
	if(!cur) return 0;

	if(!strcasecmp(cur->name, "DetectTime")) {
		if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_time()\n",cur->name);
		alert->detecttime = (IDMEFtime *)malloc(sizeof(IDMEFtime));
		memset(alert->detecttime, '\0', sizeof(IDMEFtime));
		if(!parse_time(cur, alert->detecttime, "DetectTime")) {
			fprintf(stderr, "error parsing DetectTime\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 0;
	}

	if(!strcasecmp(cur->name, "AnalyzerTime")) {
		if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_time()\n",cur->name);
		alert->analyzertime = (IDMEFtime *)malloc(sizeof(IDMEFtime));
		memset(alert->analyzertime, '\0', sizeof(IDMEFtime));
		if(!parse_time(cur, alert->analyzertime, "AnalyzerTime")) {
			fprintf(stderr, "error parsing AnalyzerTime\n");
			return 0;
		}
		cur = cur->next;
		if(!cur) return 0;
	}
		
	n = count_elements(cur, "Source");

	if(n) {
		alert->sources =
		(IDMEFsource **)malloc(n * sizeof(IDMEFsource *));
		memset(alert->sources, '\0', n * sizeof(IDMEFsource *));

		alert->nsources = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_source()\n",cur->name);
			alert->sources[i] =
			(IDMEFsource *)malloc(sizeof(IDMEFsource));
			memset(alert->sources[i], '\0', sizeof(IDMEFsource));
			if(!parse_source(cur, alert->sources[i])) {
				fprintf(stderr, "error parsing source\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	/* must return 0 before here since we didn't see Classification */
	if(!cur) return 0;

	n = count_elements(cur, "Target");

	if(n) {
		alert->targets =
		(IDMEFtarget **)malloc(n * sizeof(IDMEFtarget *));
		memset(alert->targets, '\0', n * sizeof(IDMEFtarget *));

		alert->ntargets = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_target()\n",cur->name);
			alert->targets[i] =
			(IDMEFtarget *)malloc(sizeof(IDMEFtarget));
			memset(alert->targets[i], '\0', sizeof(IDMEFtarget));
			if(!parse_target(cur, alert->targets[i])) {
				fprintf(stderr, "error parsing target\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	/* must return 0 before here since we didn't see Classification */
	if(!cur) return 0;

	/* Classification must appear here */
	if(strcasecmp(cur->name, "Classification")) return 0;

	n = count_elements(cur, "Classification");

	if(n) {
		alert->classifications =
		(IDMEFclassification **)malloc(n * sizeof(IDMEFclassification *));
		memset(alert->classifications, '\0',
		n * sizeof(IDMEFclassification *));

		alert->nclassifications = n;


		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_classification()\n",cur->name);
			alert->classifications[i] =
			(IDMEFclassification *)malloc(sizeof(IDMEFclassification));
			memset(alert->classifications[i], '\0',
			sizeof(IDMEFclassification));
			if(!parse_classification
			(cur, alert->classifications[i])) {
				fprintf(stderr,
				"error parsing Classification\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	/* can return 1 here since the rest is optional */
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "Assessment")) {
		if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_assessment()\n",cur->name);
		alert->assessment = (IDMEFassessment *)
		malloc(sizeof(IDMEFassessment));
		memset(alert->assessment, '\0', sizeof(IDMEFassessment));
		if(!parse_assessment(cur, alert->assessment)) {
			fprintf(stderr, "error parsing Assessment\n");
			return 0;
		}
		cur = cur->next;
	}

	/* can return 1 here since the rest is optional */
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "ToolAlert")) {
		if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_toolalert()\n",cur->name);
		alert->toolalert = (IDMEFtoolalert *)malloc(sizeof(IDMEFtoolalert));	
		memset(alert->toolalert,
		'\0', sizeof(IDMEFtoolalert));
		if(!parse_toolalert(cur, alert->toolalert)) {
			fprintf(stderr, "error parsing ToolAlert\n");
			return 0;
		}
		cur = cur->next;
	}

	/* can return 1 here since the rest is optional */
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "OverflowAlert")) {
		if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_overflowalert()\n",cur->name);
		alert->overflowalert = (IDMEFoverflowalert *)malloc(sizeof(IDMEFoverflowalert));
		memset(alert->overflowalert,
		'\0', sizeof(IDMEFoverflowalert));
		if(!parse_overflowalert(cur, alert->overflowalert)) {
			fprintf(stderr, "error parsing OverflowAlert\n");
			return 0;
		}
		cur = cur->next;
	}

	/* can return 1 since the rest is optional */
	if(!cur) return 1;

	if(!strcasecmp(cur->name, "CorrelationAlert")) {
		if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_correlationalert()\n",cur->name);
		alert->correlationalert = (IDMEFcorrelationalert *)malloc(sizeof(IDMEFcorrelationalert));
		memset(alert->correlationalert,
		'\0', sizeof(IDMEFcorrelationalert));
		if(!parse_correlationalert(cur, alert->correlationalert)) {
			fprintf(stderr, "error parsing CorrelationAlert\n");
			return 0;
		}
		cur = cur->next;
	}

	/* can return 1 since the rest is optional */
	if(!cur) return 1;

	/* If we get here it must be AdditionalData */
	if(strcasecmp(cur->name, "AdditionalData")) return 0;

	n = count_elements(cur, "AdditionalData");

	if(n) {
		alert->additionaldatas =
		(IDMEFadditionaldata **)malloc(n * sizeof(IDMEFadditionaldata *));
		memset(alert->additionaldatas, '\0',
		n * sizeof(IDMEFadditionaldata *));

		alert->nadditionaldatas = n;


		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_alert) next element is %s -- calling parse_additionaldata()\n",cur->name);
			alert->additionaldatas[i] =
			(IDMEFadditionaldata *)malloc(sizeof(IDMEFadditionaldata));
			memset(alert->additionaldatas[i], '\0',
			sizeof(IDMEFadditionaldata));
			if(!parse_additionaldata(cur, alert->additionaldatas[i])) {
				fprintf(stderr,
				"error parsing AdditionalData\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	return 1;
}

int
parse_message(xmlNodePtr cur, IDMEFmessage *message) {
	char *p = 0;
	int i, n;

	p = get_attribute(cur, "version",1);

	if(p) message->version = p;

	cur = cur->children;
	/* the spec is unclear about whether or not a message with */
	/* no alerts or heartbeats is legal so I say it's illegal. */
	if(!cur) return 0;

	n = count_elements(cur, "Alert");

	if(n) {
		message->alerts =
		(IDMEFalert **)malloc(n * sizeof(IDMEFalert *));
		memset(message->alerts, '\0', n * sizeof(IDMEFalert *));

		message->nalerts = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_message) next element is %s -- calling parse_alert()\n",cur->name);
			message->alerts[i] =
			(IDMEFalert *)malloc(sizeof(IDMEFalert));
			memset(message->alerts[i], '\0', sizeof(IDMEFalert));
			if(!parse_alert(cur, message->alerts[i])) {
				fprintf(stderr, "error parsing alert\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	if(!cur) return 1;

	n = count_elements(cur, "Heartbeat");

	if(n) {
		message->heartbeats =
		(IDMEFheartbeat **)malloc(n * sizeof(IDMEFheartbeat *));
		memset(message->heartbeats, '\0', n * sizeof(IDMEFheartbeat *));

		message->nheartbeats = n;

		for(i = 0; i < n && cur; i++) {
			if(DEBUG) fprintf(stdout, "(parse_message) next element is %s -- calling parse_heartbeat()\n",cur->name);
			message->heartbeats[i] =
			(IDMEFheartbeat *)malloc(sizeof(IDMEFheartbeat));
			memset(message->heartbeats[i], '\0', sizeof(IDMEFheartbeat));
			if(!parse_heartbeat(cur, message->heartbeats[i])) {
				fprintf(stderr, "error parsing heartbeat\n");
				return 0;
			}
			cur = cur->next;
		}
	}

	return 1;
}

/*
** Name: free_XXX where XXX is an element type defined in the
** IDMEF message exchange format specification.  Types are in
** lowercase in all cases.
**
** Purpose: To free all the allocated memory associated with an
** an element of type XXX.
**
** Arguments: The element which you would like free
**
** Returns: void
**
** Notes:
**
*/

void
free_webservice(IDMEFwebservice *webservice) {
	int i;

	if(webservice->url) free(webservice->url);
	if(webservice->cgi) free(webservice->cgi);
	if(webservice->http_method) free(webservice->http_method);

	for(i = 0; i < webservice->narg; i++) {
		free(webservice->arg[i]);
	}

	if(webservice->narg) free(webservice->arg);

	free(webservice);
}

void
free_snmpservice(IDMEFSNMPservice *SNMPservice) {
	if(SNMPservice->oid) free(SNMPservice->oid);
	if(SNMPservice->community) free(SNMPservice->community);
	if(SNMPservice->command) free(SNMPservice->command);

	free(SNMPservice);
}

void
free_service(IDMEFservice *service) {
	if(service->ident) free(service->ident);
	if(service->name) free(service->name);
	if(service->portlist) free(service->portlist);
	if(service->webservice) free_webservice(service->webservice);
	if(service->snmpservice) free_snmpservice(service->snmpservice);

	free(service);
}

void
free_userid(IDMEFuserid *userid) {
	if(userid->ident) free(userid->ident);
	if(userid->name) free(userid->name);

	free(userid);
}

void
free_user(IDMEFuser *user) {
	int i;

	if(user->ident) free(user->ident);

	for(i = 0; i < user->nuserids; i++) {
		free_userid(user->userids[i]);
	}

	if(user->nuserids) free(user->userids);	

	free(user);	
}

void
free_time(IDMEFtime *time) {
	if(time->ntpstamp) free(time->ntpstamp);
	if(time->string) free(time->string);

	free(time);
}

void
free_address(IDMEFaddress *address) {
	if(address->ident) free(address->ident);
	if(address->vlan_name) free(address->vlan_name);
	if(address->address) free(address->address);
	if(address->netmask) free(address->netmask);

	free(address);
}

void
free_node(IDMEFnode *node) {
	int i;

	if(node->ident) free(node->ident);
	if(node->location) free(node->location);
	if(node->name) free(node->name);

	if(node->naddresses) {
		for(i = 0; i < node->naddresses; i++) {
			free_address(node->addresses[i]);
		}
		free(node->addresses);
	}

	free(node);
}

void
free_process(IDMEFprocess *process) {
	int i;

	if(process->ident) free(process->ident);
	if(process->name) free(process->name);
	if(process->path) free(process->path);
	if(process->narg) {
		for(i = 0; i < process->narg; i++) free(process->arg[i]);
		free(process->arg);
	}
	if(process->nenv) {
		for(i = 0; i < process->nenv; i++) free(process->env[i]);
		free(process->env);
	}

	free(process);
}

void
free_fileaccess(IDMEFfileaccess *fileaccess) {
	int i;

	if(fileaccess->userid) free_userid(fileaccess->userid);

	for(i = 0; i < fileaccess->npermissions; i++) {
		free(fileaccess->permissions[i]);
	}

	if(fileaccess->npermissions) free(fileaccess->permissions);

	free(fileaccess);
}

void
free_linkage(IDMEFlinkage *linkage) {
	if(linkage->name) free(linkage->name);
	if(linkage->path) free(linkage->path);

	if(linkage->file) free_file(linkage->file);

	free(linkage);
}

void
free_inode(IDMEFinode *inode) {
	if(inode->change_time) free(inode->change_time);

	free(inode);
}

void
free_file(IDMEFfile *file) {
	int i;

	if(file->ident) free(file->ident);
	if(file->fstype) free(file->fstype);
	if(file->name) free(file->name);
	if(file->path) free(file->path);
	if(file->create_time) free(file->create_time);
	if(file->modify_time) free(file->modify_time);
	if(file->access_time) free(file->access_time);

	for(i = 0; i < file->nfileaccesses; i++) {
		free_fileaccess(file->fileaccesses[i]);
	}

	if(file->nfileaccesses) free(file->fileaccesses);	

	for(i = 0; i < file->nlinkages; i++) {
		free_linkage(file->linkages[i]);
	}

	if(file->nlinkages) free(file->linkages);	

	if(file->inode) free_inode(file->inode);

	free(file);
}

void
free_filelist(IDMEFfilelist *filelist) {
	int i;

	for(i = 0; i < filelist->nfiles; i++) {
		free_file(filelist->files[i]);
	}

	if(filelist->nfiles) free(filelist->files);
	
	free(filelist);
}

void
free_source(IDMEFsource *source) {
	if(source->ident) free(source->ident);
	if(source->interface) free(source->interface);
	if(source->node) free_node(source->node);
	if(source->user) free_user(source->user);
	if(source->process) free_process(source->process);
	if(source->service) free_service(source->service);

	free(source);
}

void
free_target(IDMEFtarget *target) {
	if(target->ident) free(target->ident);
	if(target->node) free_node(target->node);
	if(target->user) free_user(target->user);
	if(target->process) free_process(target->process);
	if(target->service) free_service(target->service);
	if(target->filelist) free_filelist(target->filelist);

	free(target);
}

void
free_analyzer(IDMEFanalyzer *analyzer) {
	if(analyzer->analyzerid) free(analyzer->analyzerid);
	if(analyzer->manufacturer) free(analyzer->manufacturer);
	if(analyzer->model) free(analyzer->model);
	if(analyzer->version) free(analyzer->version);
	if(analyzer->class) free(analyzer->class);
	if(analyzer->ostype) free(analyzer->ostype);
	if(analyzer->osversion) free(analyzer->osversion);
	if(analyzer->node) free_node(analyzer->node);
	if(analyzer->process) free_process(analyzer->process);

	free(analyzer);
}

void
free_alertident(IDMEFalertident *alertident) {
	if(alertident->data) free(alertident->data);
	if(alertident->analyzerid) free(alertident->analyzerid);

	free(alertident);
}

void
free_correlationalert(IDMEFcorrelationalert *correlationalert) {
	int i;

	if(correlationalert->name) free(correlationalert->name);

	if(correlationalert->alertidents) {
		for(i = 0; i < correlationalert->nalertidents; i++) {
			free_alertident(correlationalert->alertidents[i]);
		}
		free(correlationalert->alertidents);
	}

	free(correlationalert);
}

void
free_toolalert(IDMEFtoolalert *toolalert) {
	int i;

	if(toolalert->name) free(toolalert->name);
	if(toolalert->command) free(toolalert->command);

	if(toolalert->nalertidents) {
		for(i = 0; i < toolalert->nalertidents; i++) {
			free_alertident(toolalert->alertidents[i]);
		}
		free(toolalert->alertidents);
	}

	free(toolalert);
}

void
free_overflowalert(IDMEFoverflowalert *overflowalert) {
	if(overflowalert->program) free(overflowalert->program);
	if(overflowalert->buffer) free(overflowalert->buffer);

	free(overflowalert);
}

void
free_additionaldata(IDMEFadditionaldata *additionaldata) {
	if(additionaldata->meaning) free(additionaldata->meaning);
	if(additionaldata->data) free(additionaldata->data);

	free(additionaldata);
}

void
free_classification(IDMEFclassification *classification) {
	if(classification->name) free(classification->name);
	if(classification->url) free(classification->url);

	free(classification);
}

void
free_impact(IDMEFimpact *impact) {
	if(impact->data) free(impact->data);

	free(impact);
}

void
free_action(IDMEFaction *action) {
	if(action->data) free(action->data);

	free(action);
}

void
free_confidence(IDMEFconfidence *confidence) {
	if(confidence->data) free(confidence->data);

	free(confidence);
}

void
free_assessment(IDMEFassessment *assessment) {
	int i;

	if(assessment->impact) free_impact(assessment->impact);

	for(i = 0; i < assessment->nactions; i++)
		free_action(assessment->actions[i]);

	if(assessment->actions) free(assessment->actions);	

	if(assessment->confidence) free_confidence(assessment->confidence);

	free(assessment);
}

void
free_alert(IDMEFalert *alert) {
	int i;
	
	if(alert->ident) free(alert->ident);
	if(alert->analyzer) free_analyzer(alert->analyzer);
	if(alert->createtime) free_time(alert->createtime);
	if(alert->detecttime) free_time(alert->detecttime);
	if(alert->analyzertime) free_time(alert->analyzertime);
	
	for(i = 0; i < alert->nsources; i++) free_source(alert->sources[i]);
	
	if(alert->sources) free(alert->sources);
	
	for(i = 0; i < alert->ntargets; i++) free_target(alert->targets[i]);
	
	if(alert->targets) free(alert->targets);
	
	if(alert->nclassifications) {
		for(i = 0; i < alert->nclassifications; i++) {
			free_classification(alert->classifications[i]);
		}
		free(alert->classifications);
	}	
	
	if(alert->assessment) free_assessment(alert->assessment);

	if(alert->nadditionaldatas) {
		for(i = 0; i < alert->nadditionaldatas; i++) {
			free_additionaldata(alert->additionaldatas[i]);
		}
		free(alert->additionaldatas);
	}
	
	if(alert->correlationalert)
		free_correlationalert(alert->correlationalert);
	
	if(alert->toolalert) free_toolalert(alert->toolalert);

	if(alert->overflowalert) free_overflowalert(alert->overflowalert);

	free(alert);
}

void
free_heartbeat(IDMEFheartbeat *heartbeat) {
	int i;
	
	if(heartbeat->ident) free(heartbeat->ident);
	if(heartbeat->analyzer) free_analyzer(heartbeat->analyzer);
	if(heartbeat->createtime) free_time(heartbeat->createtime);
	if(heartbeat->analyzertime) free_time(heartbeat->analyzertime);
	
	if(heartbeat->nadditionaldatas) {
		for(i = 0; i < heartbeat->nadditionaldatas; i++) {
			free_additionaldata(heartbeat->additionaldatas[i]);
		}
		free(heartbeat->additionaldatas);
	}

	free(heartbeat);
}

void
free_message(IDMEFmessage *message) {
	int i;

	if(message->version) free(message->version);

	for(i = 0; i < message->nalerts; i++) {
		free_alert(message->alerts[i]);
	}

	if(message->alerts) free(message->alerts);

	for(i = 0; i < message->nheartbeats; i++) {
		free_heartbeat(message->heartbeats[i]);
	}

	if(message->heartbeats) free(message->heartbeats);

	free(message);
}

/*
** Name: get_idmef_message_from_file
**
** Purpose: Creates an IDMEFmessage structure from one IDMEF message
** found in a file.
**
** Arguments: The name of the file containing the message.
**
** Returns: A pointer to the static IDMEFmessage structure.
**
** Notes:
**
** 1. The pointer returned is to a static IDMEFmessage structure.
** As such each subsequent call to this function will overwrite
** the previous calls structure.
*/

IDMEFmessage *
get_idmef_message_from_file(char *filename) {
	IDMEFmessage *message;
	xmlNodePtr cur;
	static xmlDocPtr doc = 0;

	xmlKeepBlanksDefault(0);

	if(doc) xmlFreeDoc(doc);

	if(!(doc = xmlParseFile(filename))) {
		fprintf(stderr, "error: xmlParseFile()\n");
		return 0;
	}

	if(!(cur = xmlDocGetRootElement(doc))) {
		fprintf(stderr, "error: xmlDocGetRootElement()\n");
		xmlFreeDoc(doc);
		return 0;
	}
	
	if(xmlStrcmp(cur->name, "IDMEF-Message")) {
		fprintf(stderr, "not a valid IDMEF-Message\n");
		xmlFreeDoc(doc);
		return 0;
	}

	message = (IDMEFmessage *)malloc(sizeof(IDMEFmessage));
	memset(message, '\0', sizeof(IDMEFmessage));

	if(!parse_message(cur, message)) {
		fprintf(stderr, "not a valid IDMEF-Message\n");
		free_message(message);
		return 0;
	}

	return message;
}

/*
** Name: get_idmef_message
**
** Purpose: Creates an IDMEFmessage structure from one IDMEF message
** found in the given chunk of memory.
**
** Arguments: The buffer containing the message and the size of that
** buffer.
**
** Returns: A pointer to the static IDMEFmessage structure.
**
** Notes:
**
** 1. The pointer returned is to a static IDMEFmessage structure.
** As such each subsequent call to this function will overwrite
** the previous calls structure.
*/

IDMEFmessage *
get_idmef_message(char *buffer, int size) {
	IDMEFmessage *message;
	xmlNodePtr cur;
	static xmlDocPtr doc = 0;

	if(doc) xmlFreeDoc(doc);

	if(!(doc = xmlParseMemory(buffer, size))) {
		fprintf(stderr, "error: xmlParseMemory()\n");
		return 0;
	}

	if(!(cur = xmlDocGetRootElement(doc))) {
		fprintf(stderr, "error: xmlDocGetRootElement()\n");
		xmlFreeDoc(doc);
		return 0;
	}
	
	if(xmlStrcmp(cur->name, "IDMEF-Message")) {
		fprintf(stderr, "not a valid IDMEF-Message\n");
		xmlFreeDoc(doc);
		return 0;
	}

	message = (IDMEFmessage *)malloc(sizeof(IDMEFmessage));
	memset(message, '\0', sizeof(IDMEFmessage));

	if(!parse_message(cur, message)) {
		fprintf(stderr, "not a valid IDMEF-Message\n");
		free_message(message);
		return 0;
	}

	return message;
}
